from typing import Mapping

import numpy as np

from .dendrogram import Dendrogram, GroupDendrogram
from .utils import pairwise


class Deformation:
    """A helper class to handle data

    #. Split the data based on index
    #. Reorder the data based on label order
    #. Reorder the data based on cluster order
    #. Compute the ratio to split axes that match with data

    """

    is_row_split = False
    is_col_split = False
    is_row_cluster = False
    is_col_cluster = False
    _row_clustered = False
    _col_clustered = False

    data_row_reindex = None
    data_col_reindex = None

    row_breakpoints = None
    col_breakpoints = None
    row_split_order = None
    col_split_order = None

    row_dendrogram = None
    col_dendrogram = None
    row_linkage = None  # User supplied linkage
    col_linkage = None

    row_reorder_index = None
    col_reorder_index = None

    row_chunk_index = None
    col_chunk_index = None
    _use_col_meta = True
    _use_row_meta = True

    # just for storage
    wspace = 0
    hspace = 0

    row_cluster_kws = {}
    col_cluster_kws = {}

    data = None
    _nrow = None
    _ncol = None

    def __init__(self, data):
        self.set_data(data)

    def set_data(self, data):
        self.data = data
        self._nrow, self._ncol = data.shape
        self._col_clustered = False
        self._row_clustered = False

    def set_data_row_reindex(self, reindex):
        if len(reindex) != self._nrow:
            msg = (
                f"Length of reindex ({len(reindex)}) should match "
                f"data row with {self._nrow} elements"
            )
            raise ValueError(msg)
        self.data_row_reindex = reindex
        self._row_clustered = False

    def set_data_col_reindex(self, reindex):
        if len(reindex) != self._ncol:
            msg = (
                f"Length of reindex ({len(reindex)}) should match "
                f"data col with {self._ncol} elements"
            )
            raise ValueError(msg)
        self.data_col_reindex = reindex
        self._col_clustered = False

    def set_cluster(
        self,
        col=None,
        row=None,
        use_meta=True,
        linkage=None,
        meta_linkage=None,
        **kwargs,
    ):
        if col is not None:
            self.is_col_cluster = col
            self.col_cluster_kws = kwargs
            self._col_clustered = False
            self._use_col_meta = use_meta
            self.col_linkage = linkage
            self.col_meta_linkage = meta_linkage
        if row is not None:
            self.is_row_cluster = row
            self.row_cluster_kws = kwargs
            self._row_clustered = False
            self._use_row_meta = use_meta
            self.row_linkage = linkage
            self.row_meta_linkage = meta_linkage

    def get_data(self):
        data = self.data
        if self.data_row_reindex is not None:
            data = data[self.data_row_reindex]
        if self.data_col_reindex is not None:
            data = data[:, self.data_col_reindex]
        return data

    def set_split_row(self, breakpoints=None, order=None):
        if breakpoints is not None:
            self.is_row_split = True
            self.row_breakpoints = [0, *np.sort(np.asarray(breakpoints)), self._nrow]
            if order is None:
                order = np.arange(len(breakpoints) + 1)
            self.row_split_order = order

    def set_split_col(self, breakpoints=None, order=None):
        if breakpoints is not None:
            self.is_col_split = True
            self.col_breakpoints = [0, *np.sort(np.asarray(breakpoints)), self._ncol]
            if order is None:
                order = np.arange(len(breakpoints) + 1)
            self.col_split_order = order

    @property
    def row_ratios(self):
        self._run_cluster()
        if self.row_breakpoints is None:
            return None
        ratios = np.array([ix2 - ix1 for ix1, ix2 in pairwise(self.row_breakpoints)])

        if self.row_chunk_index is not None:
            return ratios[self.row_chunk_index]
        else:
            return ratios

    @property
    def col_ratios(self):
        self._run_cluster()
        if self.col_breakpoints is None:
            return None
        ratios = np.array([ix2 - ix1 for ix1, ix2 in pairwise(self.col_breakpoints)])

        if self.col_chunk_index is not None:
            return ratios[self.col_chunk_index]
        else:
            return ratios

    def set_row_chunk_order(self, order):
        self.row_chunk_index = order

    def set_col_chunk_order(self, order):
        self.col_chunk_index = order

    def split_by_row(self, data: np.ndarray):
        if not self.is_row_split:
            return data
        return [data[ix1:ix2] for ix1, ix2 in pairwise(self.row_breakpoints)]

    def split_by_col(self, data: np.ndarray):
        if not self.is_col_split:
            return data
        if data.ndim == 1:
            return [data[ix1:ix2] for ix1, ix2 in pairwise(self.col_breakpoints)]
        else:
            return [data[:, ix1:ix2] for ix1, ix2 in pairwise(self.col_breakpoints)]

    def split_cross(self, data: np.ndarray):
        if self.is_col_split & self.is_row_split:
            split_data = []
            for ix1, ix2 in pairwise(self.row_breakpoints):
                row = []
                for iy1, iy2 in pairwise(self.col_breakpoints):
                    row.append(data[ix1:ix2, iy1:iy2])
                split_data.append(row)
            return split_data
        if self.is_row_split:
            return self.split_by_row(data)
        if self.is_col_split:
            return self.split_by_col(data)
        return data

    _linkage_check_msg = (
        "If you want to specific linkage when splitting, "
        "it must be a dict-like object, "
        "with keys as group names and values as linkage"
    )

    def cluster_row(self):
        row_data = self.split_by_row(self.get_data())
        if self.is_row_split:
            if not (
                isinstance(self.row_linkage, Mapping) or (self.row_linkage is None)
            ):
                raise TypeError(self._linkage_check_msg)
            dens = []
            for chunk, k in zip(row_data, self.row_split_order):
                linkage = None
                if self.row_linkage is not None:
                    linkage = self.row_linkage.get(k)
                    if linkage is None:
                        raise KeyError(f"Linkage for group {k} is not specified")
                dens.append(
                    Dendrogram(chunk, linkage=linkage, key=k, **self.row_cluster_kws)
                )

            dg = GroupDendrogram(
                dens, linkage=self.row_meta_linkage, **self.row_cluster_kws
            )
            if self._use_row_meta:
                self.row_chunk_index = dg.reorder_index
            else:
                self.row_chunk_index = np.arange(len(dens))
            self.row_reorder_index = [d.reorder_index for d in dens]
        else:
            dg = Dendrogram(row_data, linkage=self.row_linkage, **self.row_cluster_kws)
            self.row_reorder_index = dg.reorder_index
        self.row_dendrogram = dg

    def cluster_col(self):
        col_data = self.split_by_col(self.get_data())
        if self.is_col_split:
            if not (
                isinstance(self.col_linkage, Mapping) or (self.col_linkage is None)
            ):
                raise TypeError(self._linkage_check_msg)
            dens = []
            for chunk, k in zip(col_data, self.col_split_order):
                linkage = None
                if self.col_linkage is not None:
                    linkage = self.col_linkage.get(k)
                    if linkage is None:
                        raise KeyError(f"Linkage for group {k} is not specified")
                dens.append(
                    Dendrogram(chunk.T, linkage=linkage, key=k, **self.col_cluster_kws)
                )
            dg = GroupDendrogram(
                dens, linkage=self.col_meta_linkage, **self.col_cluster_kws
            )
            if self._use_col_meta:
                self.col_chunk_index = dg.reorder_index
            else:
                self.col_chunk_index = np.arange(len(dens))
            self.col_reorder_index = [d.reorder_index for d in dens]
        else:
            dg = Dendrogram(
                col_data.T, linkage=self.col_linkage, **self.col_cluster_kws
            )
            self.col_reorder_index = dg.reorder_index
        self.col_dendrogram = dg

    def _run_cluster(self):
        """Calculation of dendrogram is expensive,
        so only calculated once"""
        if self.is_row_cluster & (not self._row_clustered):
            self.cluster_row()
            self._row_clustered = True
        if self.is_col_cluster & (not self._col_clustered):
            self.cluster_col()
            self._col_clustered = True

    def reorder_by_row(self, data, split="2d"):
        self._run_cluster()
        # no cluster, return immediately
        if not self.is_row_cluster:
            return data

        if split == "2d":
            if self.is_row_split & self.is_col_split:
                for row, order in zip(data, self.row_reorder_index):
                    for ix in range(len(row)):
                        row[ix] = row[ix][order]
                return [data[ix] for ix in self.row_chunk_index]

        if self.is_row_split:
            for ix, order in zip(range(len(data)), self.row_reorder_index):
                data[ix] = data[ix][order]
            return [data[ix] for ix in self.row_chunk_index]
        else:
            if (split == "2d") & self.is_col_split:
                return [d[self.row_reorder_index] for d in data]
            return data[self.row_reorder_index]

    def reorder_by_col(self, data, split="2d"):
        self._run_cluster()
        # no cluster, return immediately
        if not self.is_col_cluster:
            return data

        if split == "2d":
            if self.is_row_split & self.is_col_split:
                final_data = []
                for row in data:
                    for ix, order in zip(range(len(row)), self.col_reorder_index):
                        if row[ix].ndim == 2:
                            row[ix] = row[ix][:, order]
                        else:
                            row[ix] = row[ix][order]
                    final_data.append([row[ix] for ix in self.col_chunk_index])
                return final_data
            elif self.is_col_split:
                for ix, order in zip(range(len(data)), self.col_reorder_index):
                    data[ix] = data[ix][:, order]

                return [data[ix] for ix in self.col_chunk_index]
            elif self.is_split:
                return [d[:, self.col_reorder_index] for d in data]
            else:
                return data[:, self.col_reorder_index]
        # 1d list situation
        else:
            if self.is_col_split:
                for ix, order in zip(range(len(data)), self.col_reorder_index):
                    if data[ix].ndim == 2:
                        data[ix] = data[ix][:, order]
                    else:
                        data[ix] = data[ix][order]
                return [data[ix] for ix in self.col_chunk_index]
            else:
                if data.ndim == 2:
                    return data[:, self.col_reorder_index]
                else:
                    return data[self.col_reorder_index]

    def transform(self, data: np.ndarray):
        """data must be 2d array with the same shape as cluster data"""
        if not data.shape == (self._nrow, self._ncol):
            msg = (
                f"The shape of input data {data.shape} does not align with"
                f" the shape of cluster data {(self._nrow, self._ncol)}"
            )
            raise ValueError(msg)
        if self.data_row_reindex is not None:
            data = data[self.data_row_reindex]
        if self.data_col_reindex is not None:
            data = data[:, self.data_col_reindex]
        trans_data = self.split_cross(data)
        trans_data = self.reorder_by_row(trans_data, split="2d")
        trans_data = self.reorder_by_col(trans_data, split="2d")
        flatten_data = []
        if self.is_row_split & self.is_col_split:
            for chunk in trans_data:
                flatten_data += chunk
            return flatten_data
        return trans_data

    def transform_row(self, data: np.ndarray):
        data = data.T
        if data.ndim == 1:
            assert len(data) == self._nrow
        else:
            assert data.shape[0] == self._nrow

        if self.data_row_reindex is not None:
            data = data[self.data_row_reindex]

        trans_data = self.split_by_row(data)
        trans_data = self.reorder_by_row(trans_data, split="1d")
        if isinstance(trans_data, np.ndarray):
            return trans_data.T
        else:
            return [d.T for d in trans_data]

    def transform_col(self, data: np.ndarray):
        if data.ndim == 1:
            assert len(data) == self._ncol
        else:
            assert data.shape[1] == self._ncol

        if self.data_col_reindex is not None:
            if data.ndim == 2:
                data = data[:, self.data_col_reindex]
            else:
                data = data[self.data_col_reindex]

        trans_data = self.split_by_col(data)
        trans_data = self.reorder_by_col(trans_data, split="1d")
        return trans_data

    def get_row_dendrogram(self):
        # Update the cluster result
        self._run_cluster()
        return self.row_dendrogram

    def get_col_dendrogram(self):
        self._run_cluster()
        return self.col_dendrogram

    def get_row_linkage(self):
        if self.row_dendrogram is not None:
            if self.is_row_split:
                return {x.key: x.Z for x in self.row_dendrogram.orig_dens}
            else:
                return self.row_dendrogram.Z

    def get_col_linkage(self):
        if self.col_dendrogram is not None:
            if self.is_col_split:
                return {x.key: x.Z for x in self.col_dendrogram.orig_dens}
            else:
                return self.col_dendrogram.Z

    @property
    def is_split(self):
        return self.is_row_split | self.is_col_split

    @property
    def is_cluster(self):
        return self.is_row_cluster | self.is_col_cluster



from __future__ import annotations

import warnings
from copy import deepcopy
from numbers import Number
from typing import List, Dict
from uuid import uuid4

import numpy as np
from legendkit.layout import vstack, hstack
from matplotlib import pyplot as plt
from matplotlib.artist import Artist
from matplotlib.colors import is_color_like
from matplotlib.figure import Figure

from ._deform import Deformation
from .dendrogram import Dendrogram
from .exceptions import SplitTwice, DuplicatePlotter
from .layout import CrossLayout, CompositeCrossLayout, StackCrossLayout
from .plotter import RenderPlan, Title, SizedMesh
from .utils import pairwise, batched, get_plot_name, _check_side


def reorder_index(arr, order=None):
    uniq = set(arr)
    indices = {x: [] for x in uniq}
    for ix, a in enumerate(arr):
        indices[a].append(ix)

    if order is None:
        order = sorted(uniq)

    final_index = []
    for it in order:
        final_index += indices[it]
    return final_index, order


def get_breakpoints(arr):
    breakpoints = []
    for ix, (a, b) in enumerate(pairwise(arr)):
        if a != b:
            breakpoints.append(ix + 1)
    return breakpoints


class LegendMaker:
    """The factory class to handle legends"""

    layout: CrossLayout | CompositeCrossLayout | StackCrossLayout
    _legend_box: List[Artist] = None
    _legend_name: str = None

    def __init__(self) -> None:
        self._legend_grid_kws: Dict = {}
        self._legend_draw_kws: Dict = {}
        self._user_legends = {}
        self._draw_legend: bool = False

    def get_legends(self) -> Dict:
        """To get legends in a dict

        Returns
        -------
        A dict of {name: legends}

        """
        raise NotImplementedError("Should be implemented in derived class")

    def custom_legend(self, legend_func, name=None):
        """Add a custom legend

        Parameters
        ----------

        legend_func : Callable
            A function that return the legend object,
            the legend must be an `Artist <matplotlib.artist.Artists>`
        name : str, optional
            The name of the legend

        """
        if name is None:
            name = str(uuid4())
        self._user_legends[name] = legend_func

    def add_legends(
        self,
        side="right",
        pad=0.0,
        order=None,
        stack_by=None,
        stack_size=3,
        align_legends=None,
        align_stacks=None,
        legend_spacing=10,
        stack_spacing=10,
        box_padding=2,
    ):
        """Draw legend based on the order of annotation

        .. note::
            If you want to concatenate plots, please add legend after
            concatenation, this will merge legends from every plots

        Stack is a pack of legends

        Parameters
        ----------
        side : {'right', 'left', 'top', 'bottom'}, default: 'right'
            Which side to draw legend
        pad : number, default: 0
            The padding of the legend in inches
        order : array of plot name
            The order of the legend, if None, the order will be the same as the order when adding plotters.
            You need to set name for each plotter when adding them, and specify the order here.
        stack_by : {'row', 'col'}
            The direction to stack legends
        stack_size : int, default: 3
            The number of legends in a stack
        align_legends : {'left', 'right', 'top', 'bottom'}
            The side to align legends in a stack
        align_stacks : {'left', 'right', 'top', 'bottom'}
            The side to align stacks
        legend_spacing : float, default: 10
            The space between legends
        stack_spacing : float, default: 10
            The space between stacks
        box_padding : float, default: 2
            Add pad around the whole legend box

        """
        # TODO: Allow user to control where to add legends,
        #       relative to the main canvas or the whole figure
        # TODO: Allow user to add stack_size as a list
        #       Each stack can contain different number of legends
        _check_side(side)
        self._draw_legend = True
        if stack_by is None:
            stack_by = "col" if side in ["right", "left"] else "row"
        if align_stacks is None:
            align_stacks = "baseline"
        if align_legends is None:
            align_legends = "left" if stack_by == "col" else "bottom"

        self._legend_grid_kws = dict(side=side, size=0.01, pad=pad)
        self._legend_draw_kws = dict(
            order=order,
            stack_by=stack_by,
            stack_size=stack_size,
            align_legends=align_legends,
            align_stacks=align_stacks,
            legend_spacing=legend_spacing,
            stack_spacing=stack_spacing,
            box_padding=box_padding,
        )

    def remove_legends(self):
        self._draw_legend = False
        self.layout.remove_legend_ax()

    def _legends_drawer(self, ax):
        user_legends = {k: [v()] for k, v in self._user_legends.items()}
        legends = {**self.get_legends(), **user_legends}

        # force to remove all legends before drawing
        # In case some legends are added implicitly
        # This may not be a good solution
        for _, legs in legends.items():
            for leg in legs:
                try:
                    # Try to detach legend from figure
                    leg.remove()
                    # For matplotlib >= 3.10.0
                    if hasattr(leg, "_parent_figure"):
                        setattr(leg, "_parent_figure", None)
                    # For matplotlib < 3.10.0
                    if hasattr(leg, "figure"):
                        setattr(leg, "figure", None)
                except Exception:
                    pass

        legend_order = self._legend_draw_kws["order"]
        stack_by = self._legend_draw_kws["stack_by"]
        stack_size = self._legend_draw_kws["stack_size"]
        align_legends = self._legend_draw_kws["align_legends"]
        align_stacks = self._legend_draw_kws["align_stacks"]
        legend_spacing = self._legend_draw_kws["legend_spacing"]
        stack_spacing = self._legend_draw_kws["stack_spacing"]
        box_padding = self._legend_draw_kws["box_padding"]

        inner, outer = vstack, hstack
        if stack_by == "row":
            inner, outer = outer, inner

        all_legs = []
        if legend_order is None:
            for name, legs in legends.items():
                all_legs += legs
        else:
            for name in legend_order:
                all_legs += legends[name]

        bboxes = []
        for legs in batched(all_legs, stack_size):
            box = inner(legs, align=align_legends, spacing=legend_spacing)
            bboxes.append(box)
        legend_box = outer(
            bboxes,
            align=align_stacks,
            loc="center left",
            spacing=stack_spacing,
            padding=box_padding,
        )
        ax.add_artist(legend_box)
        # uncomment this to visualize legend ax
        # from matplotlib.patches import Rectangle
        # rect = Rectangle((0, 0), 1, 1,
        #                  fill=None,
        #                  edgecolor="r",
        #                  transform=ax.transAxes)
        # ax.add_artist(rect)
        return legend_box

    def _freeze_legend(self, figure):
        if self._draw_legend:
            if self.layout.get_legend_ax() is None:
                self.layout.add_legend_ax(**self._legend_grid_kws)
            renderer = figure.canvas.get_renderer()
            legend_ax = figure.add_axes([0, 0, 1, 1])
            legends_box = self._legends_drawer(legend_ax)
            bbox = legends_box.get_window_extent(renderer)
            if self._legend_grid_kws["side"] in ["left", "right"]:
                size = bbox.xmax - bbox.xmin
            else:
                size = bbox.ymax - bbox.ymin
            self.layout.set_legend_size(size / figure.get_dpi())
            legend_ax.remove()

    def _render_legend(self):
        if self._draw_legend:
            legend_ax = self.layout.get_legend_ax()
            legend_ax.set_axis_off()
            legend_box = self._legends_drawer(legend_ax)
            self._legend_box = legend_box


class WhiteBoard(LegendMaker):
    """The base class that handle all rendering process

    Parameters
    ----------
    width : float, optional
        The width of the main canvas in inches
    height : float, optional
        The height of the main canvas in inches
    name : str, optional
        The name of the main canvas
    margin : float, 4-tuple, optional
        The margin of the main canvas in inches
    init_main : bool, optional
        If True, the main canvas will be initialized


    See Also
    --------
    :class:`~marsilea.base.ClusterBoard`


    Attributes
    ----------
    layout : CrossLayout
        The layout manager
    figure : Figure
        The matplotlib figure object

    Examples
    --------
    Create a violin plot in white board

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> import marsilea as ma
        >>> data = np.random.rand(10, 10)
        >>> h = ma.WhiteBoard(height=2)
        >>> h.add_layer(ma.plotter.Violin(data))
        >>> h.render()


    """

    layout: CrossLayout
    figure: Figure = None
    _row_plan: List[RenderPlan]
    _col_plan: List[RenderPlan]
    _layer_plan: List[RenderPlan]

    def __init__(self, width=None, height=None, name=None, margin=0.2, init_main=True):
        self.main_name = get_plot_name(name, "main", "board")
        self._main_size_updatable = (width is None) & (height is None)
        width = 4 if width is None else width
        height = 4 if height is None else height
        self.layout = CrossLayout(
            name=self.main_name,
            width=width,
            height=height,
            margin=margin,
            init_main=init_main,
        )

        # self._side_count = {"right": 0, "left": 0, "top": 0, "bottom": 0}
        self._col_plan = []
        self._row_plan = []
        self._layer_plan = []
        # use to mark if legend is enabled for a RenderPlan
        self._legend_switch = {}
        super().__init__()

    def add_plot(
        self, side, plot: RenderPlan, name=None, size=None, pad=0.0, legend=True
    ):
        """Add a plotter to the board

        Parameters
        ----------
        side : {"left", "right", "top", "bottom"}
            Which side to add the plotter
        plot : RenderPlan
            The plotter to add
        name : str, optional
            The name of the plot
        size : float, optional
            The size of the plot in inches
        pad : float, optional
            The padding of the plot in inches
        legend : bool, optional
            If True, the legend will be included when calling :meth:`~marsilea.base.LegendMaker.add_legends`

        """
        if plot.name is not None:
            raise DuplicatePlotter(plot)
        plot_name = get_plot_name(name, side, plot.__class__.__name__)
        self._legend_switch[plot_name] = legend

        if size is not None:
            ax_size = size
        else:
            if plot.size is not None:
                ax_size = plot.size
            else:
                ax_size = 1.0

        self.layout.add_ax(side, name=plot_name, size=ax_size, pad=pad)

        if side in ["top", "bottom"]:
            plan = self._col_plan
        else:
            plan = self._row_plan
        plot.set(name=plot_name, size=size)
        plot.set_side(side)

        plan.append(plot)

    def add_left(self, plot: RenderPlan, name=None, size=None, pad=0.0, legend=True):
        """Add a plotter to the left-side of main canvas

        Parameters
        ----------
        plot : RenderPlan
            The plotter to add
        name : str, optional
            The name of the plot
        size : float, optional
            The size of the plot in inches
        pad : float, optional
            The padding of the plot in inches
        legend : bool, optional
            If True, the legend will be included when calling :meth:`~marsilea.base.LegendMaker.add_legends`

        """
        self.add_plot("left", plot, name, size, pad, legend)

    def add_right(self, plot: RenderPlan, name=None, size=None, pad=0.0, legend=True):
        """Add a plotter to the right-side of main canvas

        Parameters
        ----------
        plot : RenderPlan
            The plotter to add
        name : str, optional
            The name of the plot
        size : float, optional
            The size of the plot in inches
        pad : float, optional
            The padding of the plot in inches
        legend : bool, optional
            If True, the legend will be included when calling :meth:`~marsilea.base.LegendMaker.add_legends`

        """
        self.add_plot("right", plot, name, size, pad, legend)

    def add_top(self, plot: RenderPlan, name=None, size=None, pad=0.0, legend=True):
        """Add a plotter to the top-side of main canvas

        Parameters
        ----------
        plot : RenderPlan
            The plotter to add
        name : str, optional
            The name of the plot
        size : float, optional
            The size of the plot in inches
        pad : float, optional
            The padding of the plot in inches
        legend : bool, optional
            If True, the legend will be included when calling :meth:`~marsilea.base.LegendMaker.add_legends`

        """
        self.add_plot("top", plot, name, size, pad, legend)

    def add_bottom(self, plot: RenderPlan, name=None, size=None, pad=0.0, legend=True):
        """Add a plotter to the bottom-side of main canvas

        Parameters
        ----------
        plot : RenderPlan
            The plotter to add
        name : str, optional
            The name of the plot
        size : float, optional
            The size of the plot in inches
        pad : float, optional
            The padding of the plot in inches
        legend : bool, optional
            If True, the legend will be included when calling :meth:`~marsilea.base.LegendMaker.add_legends`

        """
        self.add_plot("bottom", plot, name, size, pad, legend)

    def _render_plan(self):
        for plan in self._col_plan:
            axes = self.layout.get_ax(plan.name)
            plan.render(axes)

        # render other plots
        for plan in self._row_plan:
            axes = self.layout.get_ax(plan.name)
            plan.render(axes)

        main_ax = self.get_main_ax()
        for plan in self._get_layers_zorder():
            plan.render(main_ax)

    def add_layer(self, plot: RenderPlan, zorder=None, name=None, legend=True):
        """Add a plotter to the main canvas

        .. note::

            Not every plotter can be added as a layer.

        Parameters
        ----------
        plot : RenderPlan
            The plotter to add
        zorder : int, optional
            The zorder of the plot
        name : str, optional
            The name of the plot
        legend : bool, optional
            If True, the legend will be included when calling :meth:`~marsilea.base.LegendMaker.add_legends`

        """
        if name is None:
            name = plot.name
        plot_type = plot.__class__.__name__
        name = get_plot_name(name, side="main", chart=plot_type)
        self._legend_switch[name] = legend
        if not plot.render_main:
            msg = f"{plot_type} " f"cannot be rendered as another layer."
            raise TypeError(msg)
        if zorder is not None:
            plot.zorder = zorder
        plot.set(name=name)
        plot.set_side("main")
        self._layer_plan.append(plot)

        # SizedMesh will update the main canvas size
        if self._main_size_updatable:
            if isinstance(plot, SizedMesh):
                w, h = plot.update_main_canvas_size()
                self.layout.set_main_width(w)
                self.layout.set_main_height(h)
                # only update once,
                # if we have more plot in the future
                # that will change canvas size
                self._main_size_updatable = False

    def _get_layers_zorder(self):
        return sorted(self._layer_plan, key=lambda p: p.zorder)

    def add_pad(self, side, size):
        """Add padding to the main canvas

        Parameters
        ----------
        side : {"left", "right", "top", "bottom"}
            Which side to add padding
        size : float
            The size of padding in inches

        """
        self.layout.add_pad(side, size)

    def add_canvas(self, side, name, size, pad=0.0):
        """Add an axes to the main canvas

        Parameters
        ----------
        side : {"left", "right", "top", "bottom"}
            Which side to add the axes
        name : str
            The name of the axes
        size : float
            The size of the axes in inches
        pad : float, optional
            The padding of the axes in inches

        """
        self.layout.add_ax(side, name, size, pad=pad)

    def add_title(self, top=None, bottom=None, left=None, right=None, pad=0, **props):
        """A shortcut to add title to the main canvas

        Parameters
        ----------
        top : str, optional
            The title of the top side
        bottom : str, optional
            The title of the bottom side
        left : str, optional
            The title of the left side
        right : str, optional
            The title of the right side
        pad : float, optional
            The padding of the title in inches
        props : dict
            The properties of the title

        Returns
        -------

        """
        if left is not None:
            self.add_plot("left", Title(left, **props), pad=pad)
        if right is not None:
            self.add_plot("right", Title(right, **props), pad=pad)
        if top is not None:
            self.add_plot("top", Title(top, **props), pad=pad)
        if bottom is not None:
            self.add_plot("bottom", Title(bottom, **props), pad=pad)

    def get_ax(self, name):
        """Get a specific axes by name when available

        If the axes is split, multiple axes will be returned

        .. note::
            This will not work before `render` is called

        """
        return self.layout.get_ax(name)

    def get_main_ax(self):
        """Return the main axes"""
        return self.layout.get_main_ax()

    def _extra_legends(self):
        """If there are legends that cannot get from RenderPlan

        Must be overridden in derived class

        """
        return {}

    def get_legends(self):
        """Get all legends from the main canvas"""
        legends = {}
        legends.update(self._extra_legends())
        for plan in self._layer_plan + self._col_plan + self._row_plan:
            # Not every render plan has legend
            if self._legend_switch[plan.name]:
                legs = plan.get_legends()
                if legs is not None:
                    if isinstance(legs, Artist):
                        legs = [legs]
                    legends[plan.name] = legs
        return legends

    def __add__(self, other):
        """Define behavior that horizontal appends two grid"""
        return self.append("right", other)

    def __truediv__(self, other):
        """Define behavior that vertical appends two grid"""
        return self.append("bottom", other)

    def append(self, side, other):
        """Append two :class:`~marsilea.base.CrossLayout` together"""
        compose_board = CompositeBoard(self)
        compose_board.append(side, other)
        return compose_board

    def _freeze_flex_plots(self, figure):
        main_cell = self.layout.main_cell
        main_width = main_cell.width
        main_height = main_cell.height

        for plan in self._col_plan + self._row_plan:
            if plan.size is None:
                render_size = plan.get_canvas_size(
                    figure, main_width=main_width, main_height=main_height
                )
                if render_size is not None:
                    self.layout.set_render_size(plan.name, render_size)

    def render(self, figure=None, scale=1):
        """Finalize the layout and render all plots

        Parameters
        ----------
        figure : :class:`~matplotlib.figure.FigureBase`, optional
            The matplotlib figure object
        scale : float, optional
            The scale value of the figure size. You can use this to
            adjust the overall size of the figure

        Returns
        -------

        """
        if figure is None:
            figure = plt.figure()
        self.figure = figure
        self._freeze_legend(figure)
        self._freeze_flex_plots(figure)

        self.layout.freeze(figure=figure, scale=scale)

        # render other plots
        self._render_plan()
        self._render_legend()

    def save(self, fname, **kwargs):
        """Save the figure to a file

        This will force a re-render of the figure

        Parameters
        ----------
        fname : str, path-like
            The file name to save
        kwargs : dict
            Additional options for saving the figure, will be passed to :meth:`~matplotlib.pyplot.savefig`

        """
        self.render()
        save_options = dict(bbox_inches="tight")
        save_options.update(kwargs)
        self.figure.savefig(fname, **save_options)

    def set_margin(self, margin: float | tuple[float, float, float, float]):
        """Set margin of the main canvas

        Parameters
        ----------
        margin : float, 4-tuple
            The margin of the main canvas in inches

        """
        self.layout.set_margin(margin)


class ZeroWidth(WhiteBoard):
    """A utility class to initialize a canvas \
    with zero width

    This is useful when you try to stack many plots

    Parameters
    ----------
    height : float
        The
    name : str
    margin : float

    """

    def __init__(self, height, name=None, margin=0.2):
        super().__init__(
            width=0, height=height, name=name, margin=margin, init_main=False
        )


class ZeroHeight(WhiteBoard):
    """A utility class to initialize a canvas \
    with zero height

    This is useful when you try to stack many plots

    """

    def __init__(self, width, name=None, margin=0.2):
        super().__init__(
            width=width, height=0, name=name, margin=margin, init_main=False
        )


class CompositeBoard(LegendMaker):
    """Layout multiple canvas

    Parameters
    ----------
    main_board : :class:`WhiteBoard` or :class:`ClusterBoard`
        The main canvas
    keep_legends : bool, default: False
        Whether to keep the legends in each canvas
        If False, you can group all legends with `.add_legends()`
    align_main : bool, default: True
        Whether to force the size of other canvas to align with the main canvas
    margin : float, default: 0
        The margin space reserved around the whole canvas

    """

    layout: CompositeCrossLayout = None
    figure: Figure = None

    def __init__(
        self,
        main_board: WhiteBoard,
        keep_legends=False,
        align_main=True,
        margin=0,
    ):
        self.keep_legends = keep_legends

        self.main_board = self.new_board(main_board)
        if not keep_legends:
            self.main_board.remove_legends()
        self.layout = CompositeCrossLayout(
            self.main_board.layout, align_main=align_main, margin=margin
        )
        self._board_list = [self.main_board]

        super().__init__()

    def new_board(self, board):
        board = deepcopy(board)
        if not self.keep_legends & isinstance(board, LegendMaker):
            board.remove_legends()
        return board

    def __add__(self, other):
        """Define behavior that horizontal appends two grid"""
        return self.append("right", other)

    def __truediv__(self, other):
        """Define behavior that vertical appends two grid"""
        return self.append("bottom", other)

    def append(self, side, other, pad=0):
        if isinstance(other, Number):
            self.layout.append(side, other)
        else:
            board = self.new_board(other)
            self._board_list.append(board)
            self.layout.append(side, board.layout)

        if pad > 0:
            self.layout.append(side, pad)
        return self

    def render(self, figure=None, scale=1):
        if figure is None:
            figure = plt.figure()
        self._freeze_legend(figure)
        for board in self._board_list:
            board._freeze_flex_plots(figure)
        self.layout.freeze(figure=figure, scale=scale)
        self.figure = figure
        for board in self._board_list:
            board.render(figure=self.figure)

        self._render_legend()

    def save(self, fname, **kwargs):
        if self.figure is not None:
            save_options = dict(bbox_inches="tight")
            save_options.update(kwargs)
            self.figure.savefig(fname, **save_options)
        else:
            warnings.warn(
                "Figure does not exist, " "please render it before saving as file."
            )

    def get_legends(self):
        legends = {}
        for m in self._board_list:
            legends.update(m.get_legends())
        return legends

    def get_ax(self, board_name, ax_name):
        return self.layout.get_ax(board_name, ax_name)

    def get_main_ax(self, name):
        return self.layout.get_main_ax(name)

    def set_margin(self, margin):
        self.layout.set_margin(margin)


class StackBoard(LegendMaker):
    """Stack multiple boards

    Parameters
    ----------
    boards : list of :class:`WhiteBoard`, :class:`StackBoard`

    """

    def __init__(
        self,
        boards: List[WhiteBoard, StackBoard],
        direction="horizontal",
        align="center",
        spacing=0.2,
        margin=0,
        keep_legends=False,
    ):
        self.keep_legends = keep_legends
        board_list = []
        layouts = []
        for board in boards:
            board = self.new_board(board)
            board_list.append(board)
            layouts.append(board.layout)

        self.layout = StackCrossLayout(
            layouts, margin=margin, direction=direction, align=align, spacing=spacing
        )

        self._board_list = board_list
        super().__init__()

    # To mimic the board API
    def _freeze_flex_plots(self, figure):
        for board in self._board_list:
            board._freeze_flex_plots(figure)

    def new_board(self, board):
        board = deepcopy(board)
        if not self.keep_legends & isinstance(board, LegendMaker):
            board.remove_legends()
        return board

    def render(self, figure=None, scale=1):
        if figure is None:
            figure = plt.figure()
        self._freeze_legend(figure)
        for board in self._board_list:
            board._freeze_flex_plots(figure)
        self.layout.freeze(figure=figure, scale=scale)
        self.figure = figure
        for board in self._board_list:
            board.render(figure=self.figure)

        self._render_legend()

    def save(self, fname, **kwargs):
        if self.figure is not None:
            save_options = dict(bbox_inches="tight")
            save_options.update(kwargs)
            self.figure.savefig(fname, **save_options)
        else:
            warnings.warn(
                "Figure does not exist, " "please render it before saving as file."
            )

    def get_legends(self):
        legends = {}
        for m in self._board_list:
            legends.update(m.get_legends())
        return legends

    def get_ax(self, board_name, ax_name):
        return self.layout.get_ax(board_name, ax_name)

    def get_main_ax(self, name):
        return self.layout.get_main_ax(name)

    def set_margin(self, margin):
        self.layout.set_margin(margin)


class ClusterBoard(WhiteBoard):
    """A main canvas class that can handle cluster data

    Parameters
    ----------
    cluster_data : ndarray
        The cluster data
    width : float, optional
        The width of the main canvas in inches
    height : float, optional
        The height of the main canvas in inches
    name : str, optional
        The name of the main canvas
    margin : float, 4-tuple, optional
        The margin of the main canvas in inches
    init_main : bool, optional
        If True, the main canvas will be initialized


    See Also
    --------
    :class:`~marsilea.base.WhiteBoard`


    """

    _row_reindex: List[int] = None
    _col_reindex: List[int] = None
    # If cluster data need to be defined by user
    _allow_cluster: bool = True
    _split_col: bool = False
    _split_row: bool = False
    _mesh = None

    def __init__(
        self,
        cluster_data,
        width=None,
        height=None,
        name=None,
        margin=0.2,
        init_main=True,
    ):
        super().__init__(
            width=width, height=height, name=name, margin=margin, init_main=init_main
        )
        self._row_den = []
        self._col_den = []
        cluster_data = np.asarray(cluster_data)
        if cluster_data.ndim != 2:
            raise ValueError("Cluster data must be 2D array")
        self._cluster_data = cluster_data
        self._deform = Deformation(cluster_data)

    def add_dendrogram(
        self,
        side,
        method=None,
        metric=None,
        linkage=None,
        meta_linkage=None,
        add_meta=True,
        add_base=True,
        add_divider=True,
        meta_color=None,
        linewidth=None,
        colors=None,
        divider_style="--",
        meta_ratio=0.2,
        show=True,
        name=None,
        size=0.5,
        pad=0.0,
        get_meta_center=None,
        rasterized=False,
    ):
        """Run cluster and add dendrogram

        .. note::

            #. method and metric only works when you
               add the first row/col dendrogram.
            #. If `add_meta=False` and `add_base=False`, the dendrogram
               axes will not be created.

        Parameters
        ----------
        side
        method : str
            See scipy's :meth:`linkage <scipy.cluster.hierarchy.linkage>`
        metric : str
            See scipy's :meth:`linkage <scipy.cluster.hierarchy.linkage>`
        linkage : ndarray
            Precomputed linkage matrix.
            See scipy's :meth:`linkage <scipy.cluster.hierarchy.linkage>` for
            specific format.
        meta_linkage : ndarray
            Precomputed chunk-level linkage matrix.
            See scipy's :meth:`linkage <scipy.cluster.hierarchy.linkage>` for
            specific format.
        add_meta : None | bool
            By default, add_meta is set to False if the linkage is provided, otherwise True.
            If the data is split, a meta dendrogram can be drawn for data
            chunks. The mean value of the data chunk is used to calculate
            linkage matrix for meta dendrogram.
        add_base : None | bool
            By default, add_meta is set to False if the linkage is provided, otherwise True.
            Draw the base dendrogram for each data chunk. You can turn this
            off if the base dendrogram is too crowded.
        add_divider : bool
            Draw a divide line that tells the difference between
            base and meta dendrogram.
        divider_style : str, default: "--"
            The line style of the divide line
        meta_color : color
            The color of the meta dendrogram
        meta_ratio : float
            The size of meta dendrogram relative to the base dendrogram
        linewidth : float
            The linewidth for every dendrogram and divide line
        colors : color, array of color
            If only one color is specified, it will be applied to
            every dendrogram. If an array of color is specified, it will
            be applied to each base dendrogram.
        show : bool
            If False, the dendrogram will not be drawn and the axes will
            not be created.
        name : str
            The name of the dendrogram axes
        size : float
        pad : float
        get_meta_center: callable
            A function to calculate the centroid of data. It should accept a 2D numpy
            array as input and return a 1D numpy array of the same length as the number
            of columns in the input, representing the centroid. The default will use the
            mean values.
        rasterized : bool
            If True, the dendrogram will be rasterized

        Examples
        --------

        You can change how the linkage matrix is calculated

        .. plot::
            :context: close-figs

            >>> data = np.random.rand(10, 11)
            >>> import marsilea as ma
            >>> h = ma.Heatmap(data)
            >>> h.add_dendrogram("left", method="ward", colors="green")
            >>> h.render()

        Only show the meta dendrogram to avoid crowded dendrogram

        .. plot::
            :context: close-figs

            >>> h = ma.Heatmap(data)
            >>> h.cut_rows(cut=[4, 8])
            >>> h.add_dendrogram("left", add_base=False)
            >>> h.render()

        Change color for each base dendrogram

        .. plot::
            :context: close-figs

            >>> h = ma.Heatmap(data)
            >>> h.cut_rows(cut=[4, 8])
            >>> h.add_dendrogram("left", colors=["#5470c6", "#91cc75", "#fac858"])
            >>> h.render()

        """
        if not self._allow_cluster:
            msg = (
                f"Please specify cluster data when initialize "
                f"'{self.__class__.__name__}' class."
            )
            raise ValueError(msg)
        plot_name = get_plot_name(name, side, "Dendrogram")

        # if only colors is passed
        # the color should be applied to all
        if (colors is not None) & (is_color_like(colors)) & (meta_color is None):
            meta_color = colors

        # if nothing is added
        # close the dendrogram
        if (not add_meta) & (not add_base):
            show = False

        if show:
            self.layout.add_ax(side, name=plot_name, size=size, pad=pad)

        den_options = dict(
            name=plot_name,
            show=show,
            side=side,
            add_meta=add_meta,
            add_base=add_base,
            add_divider=add_divider,
            meta_color=meta_color,
            linewidth=linewidth,
            colors=colors,
            divider_style=divider_style,
            meta_ratio=meta_ratio,
            rasterized=rasterized,
        )

        deform = self.get_deform()
        if side in ["right", "left"]:
            den_options["pos"] = "row"
            self._row_den.append(den_options)
            deform.set_cluster(
                row=True,
                method=method,
                metric=metric,
                linkage=linkage,
                meta_linkage=meta_linkage,
                use_meta=add_meta,
                get_meta_center=get_meta_center,
                rasterized=rasterized,
            )
        else:
            den_options["pos"] = "col"
            self._col_den.append(den_options)
            deform.set_cluster(
                col=True,
                method=method,
                metric=metric,
                linkage=linkage,
                meta_linkage=meta_linkage,
                use_meta=add_meta,
                get_meta_center=get_meta_center,
                rasterized=rasterized,
            )

    def group_rows(self, group, order=None, spacing=0.01):
        """Group rows into chunks

        Parameters
        ----------
        group : array-like
            The group of each row
        order : array-like, optional
            The order of the unique groups
        spacing : float, optional
            The spacing between each split chunks, default is 0.01

        Examples
        --------
        Group rows by the unique labels

        .. plot::
            :context: close-figs

            >>> data = np.random.rand(10, 11)
            >>> import marsilea as ma
            >>> h = ma.Heatmap(data)
            >>> labels = ["A", "B", "C", "A", "B", "C", "A", "B", "C", "A"]
            >>> h.group_rows(labels, order=["A", "B", "C"])
            >>> h.add_left(ma.plotter.Labels(labels), pad=0.1)
            >>> h.render()

        """
        if self._split_row:
            raise SplitTwice(axis="rows")
        self._split_row = True

        deform = self.get_deform()
        deform.hspace = spacing

        labels = np.asarray(group)
        reindex, order = reorder_index(labels, order=order)
        deform.set_data_row_reindex(reindex)

        breakpoints = get_breakpoints(labels[reindex])
        deform.set_split_row(breakpoints=breakpoints, order=order)

    def group_cols(self, group, order=None, spacing=0.01):
        """Group columns into chunks

        Parameters
        ----------
        group : array-like
            The group of each column
        order : array-like, optional
            The order of the unique groups
        spacing : float, optional
            The spacing between each split chunks, default is 0.01

        Examples
        --------
        Group columns by the unique labels

        .. plot::
            :context: close-figs

            >>> data = np.random.rand(11, 10)
            >>> import marsilea as ma
            >>> h = ma.Heatmap(data)
            >>> labels = ["A", "B", "C", "A", "B", "C", "A", "B", "C", "A"]
            >>> h.group_cols(labels, order=["A", "B", "C"])
            >>> h.add_top(ma.plotter.Labels(labels), pad=0.1)
            >>> h.render()

        """
        if self._split_col:
            raise SplitTwice(axis="columns")
        self._split_col = True

        deform = self.get_deform()
        deform.wspace = spacing

        labels = np.asarray(group)
        reindex, order = reorder_index(labels, order=order)
        deform.set_data_col_reindex(reindex)

        breakpoints = get_breakpoints(labels[reindex])
        deform.set_split_col(breakpoints=breakpoints, order=order)

    def cut_rows(self, cut, spacing=0.01):
        """Cut the main canvas by rows

        Parameters
        ----------
        cut : array-like
            The index of your data to specify where to cut the canvas
        spacing : float, optional
            The spacing between each cut, default is 0.01

        Examples
        --------
        Cut the canvas by the index

        .. plot::
            :context: close-figs

            >>> data = np.random.rand(10, 11)
            >>> import marsilea as ma
            >>> h = ma.Heatmap(data)
            >>> h.cut_rows([4, 8])
            >>> h.render()

        """
        if self._split_row:
            raise SplitTwice(axis="horizontally")
        self._split_row = True

        deform = self.get_deform()
        deform.hspace = spacing
        deform.set_split_row(breakpoints=cut)

    def cut_cols(self, cut, spacing=0.01):
        """Cut the main canvas by columns

        Parameters
        ----------
        cut : array-like
            The index of your data to specify where to cut the canvas
        spacing : float, optional
            The spacing between each cut, default is 0.01

        Examples
        --------
        Cut the canvas by the index

        .. plot::
            :context: close-figs

            >>> data = np.random.rand(10, 11)
            >>> import marsilea as ma
            >>> h = ma.Heatmap(data)
            >>> h.cut_cols([4, 8])
            >>> h.render()

        """
        if self._split_col:
            raise SplitTwice(axis="vertically")
        self._split_col = True

        deform = self.get_deform()
        deform.wspace = spacing
        deform.set_split_col(breakpoints=cut)

    def _setup_axes(self):
        deform = self.get_deform()
        w_ratios = deform.col_ratios
        h_ratios = deform.row_ratios
        wspace = deform.wspace
        hspace = deform.hspace

        # split the main axes
        if deform.is_split:
            if w_ratios is not None:
                self.layout.vsplit(self.main_name, w_ratios, wspace)
            if h_ratios is not None:
                self.layout.hsplit(self.main_name, h_ratios, hspace)

        # split column axes
        if deform.is_col_split:
            for plan in self._col_plan:
                if plan.allow_split:
                    # if deform.is_col_cluster:
                    #     group_ratios = None
                    # else:
                    group_ratios = plan.get_split_regroup()
                    self.layout.vsplit(plan.name, w_ratios, wspace, group_ratios)

        # split row axes
        if deform.is_row_split:
            for plan in self._row_plan:
                if plan.allow_split:
                    # if deform.is_row_cluster:
                    #     group_ratios = None
                    # else:
                    group_ratios = plan.get_split_regroup()
                    self.layout.hsplit(plan.name, h_ratios, hspace, group_ratios)

    def _render_dendrogram(self):
        deform = self.get_deform()
        for den in self._row_den + self._col_den:
            if den["show"]:
                ax = self.layout.get_ax(den["name"])
                ax.set_axis_off()
                spacing = deform.hspace
                den_obj = deform.get_row_dendrogram()
                if den["pos"] == "col":
                    spacing = deform.wspace
                    den_obj = deform.get_col_dendrogram()
                if isinstance(den_obj, Dendrogram):
                    color = den["colors"]
                    if (color is not None) & (not is_color_like(color)):
                        color = color[0]
                    den_obj.draw(
                        ax,
                        orient=den["side"],
                        color=color,
                        linewidth=den["linewidth"],
                        rasterized=den["rasterized"],
                    )
                else:
                    den_obj.draw(
                        ax,
                        orient=den["side"],
                        spacing=spacing,
                        add_meta=den["add_meta"],
                        add_base=den["add_base"],
                        base_colors=den["colors"],
                        meta_color=den["meta_color"],
                        linewidth=den["linewidth"],
                        divide=den["add_divider"],
                        divide_style=den["divider_style"],
                        meta_ratio=den["meta_ratio"],
                        rasterized=den["rasterized"],
                    )

    def _render_plan(self):
        deform = self.get_deform()
        for plan in self._col_plan:
            if plan.allow_split:
                plan.set_deform(deform)
            axes = self.layout.get_ax(plan.name)
            plan.render(axes)

        # render other plots
        for plan in self._row_plan:
            if plan.allow_split:
                plan.set_deform(deform)
            axes = self.layout.get_ax(plan.name)
            plan.render(axes)

        main_ax = self.get_main_ax()
        for plan in self._get_layers_zorder():
            plan.set_deform(deform)
            plan.render(main_ax)

    def get_deform(self):
        """Return the deformation object of the cluster data"""
        return self._deform

    def get_row_linkage(self):
        """Return the linkage matrix of row dendrogram

        If the canvas is not split, the linkage matrix will be returned;
        otherwise, a dictionary of linkage matrix will be returned, the key is either
        index or the name of each chunk.

        """
        return self._deform.get_row_linkage()

    def get_col_linkage(self):
        """Return the linkage matrix of column dendrogram

        If the canvas is not split, the linkage matrix will be returned;
        otherwise, a dictionary of linkage matrix will be returned, the key is either
        index or the name of each chunk.

        """
        return self._deform.get_col_linkage()

    @property
    def row_cluster(self):
        """If row dendrogram is added"""
        return len(self._row_den) > 0

    @property
    def col_cluster(self):
        """If column dendrogram is added"""
        return len(self._col_den) > 0

    def render(self, figure=None, scale=1):
        if self._deform is None:
            raise ValueError("No layer is added to the plot")
        if figure is None:
            figure = plt.figure()
        self.figure = figure
        self._freeze_legend(figure)
        self._freeze_flex_plots(figure)

        # Make sure all axes is split
        self._setup_axes()
        # Place axes
        # if refreeze:
        #     self.figure = self.layout.freeze(figure=figure, scale=scale)
        # else:
        #     self.figure = self.layout.figure
        self.layout.freeze(figure=figure, scale=scale)
        # render other plots
        self._render_plan()
        # add row and col dendrogram
        self._render_dendrogram()
        self._render_legend()


class ZeroWidthCluster(ClusterBoard):
    """
    A utility class to initialize a canvas \
    with zero width and cluster data

    Parameters
    ----------
    cluster_data : ndarray
        The cluster data
    height : float
        The height of the main canvas in inches
    name : str
        The name of the main canvas
    margin : float, 4-tuple
        The margin of the main canvas in inches

    """

    def __init__(self, cluster_data, height, name=None, margin=0.2):
        super().__init__(
            cluster_data=cluster_data,
            width=0,
            height=height,
            name=name,
            margin=margin,
            init_main=False,
        )


class ZeroHeightCluster(ClusterBoard):
    """
    A utility class to initialize a canvas \
    with zero height and cluster data

    Parameters
    ----------
    cluster_data : ndarray
        The cluster data
    width : float
        The width of the main canvas in inches
    name : str
        The name of the main canvas
    margin : float, 4-tuple
        The margin of the main canvas in inches


    """

    def __init__(self, cluster_data, width, name=None, margin=0.2):
        super().__init__(
            cluster_data=cluster_data,
            width=width,
            height=0,
            name=name,
            margin=margin,
            init_main=False,
        )



import warnings
from collections import Counter
from copy import deepcopy
from dataclasses import dataclass
from itertools import count
from typing import Any

import numpy as np
import pandas as pd
from marsilea import ClusterBoard
from marsilea.layers import LayersMesh, FrameRect, Piece
from marsilea.plotter import Labels, StackBar, Numbers, ColorMesh
from marsilea.utils import get_canvas_size_by_data

from .preset import SHAPE_BANK, MATCH_POOL, Alteration


def guess_alteration(event: str):
    for alt, rule in MATCH_POOL.items():
        if rule.is_match(event):
            return alt
    return Alteration.OTHER


@dataclass(repr=False)
class LayerData:
    matrix: np.ndarray
    piece: Piece
    color: Any
    event: Any

    def __repr__(self):
        return f"{self.piece.label} ({self.color})"


class GenomicData:
    """Handle class for genomics data

    Parameters
    ----------
    data : pd.DataFrame
        Each column is:
            1) Sample ID
            2) Track name
            3) Alteration
    samples_order : list, optional
        The order of samples, by default None
    tracks_order : list, optional
        The order of tracks, by default None
    custom_pieces : dict, optional
        Custom pieces for each alteration, by default None

    """

    def __repr__(self):
        ntrack, nsample = self.shape
        return (
            f"{ntrack} Tracks, {nsample} Samples with "
            f"{len(self.events)} Alterations"
        )

    def __init__(
        self,
        data,
        samples_order=None,
        tracks_order=None,
        custom_pieces=None,
    ):
        self.data = data.copy()
        self.data.columns = ["sample", "track", "event"]

        if samples_order is None:
            samples_order = self.data["sample"].unique()

        self.samples = samples_order
        self._patients_ix = dict(zip(samples_order, count(0, 1)))

        if tracks_order is None:
            tracks_order = self.data["track"].unique()
        self.tracks = tracks_order
        self._tracks_ix = dict(zip(tracks_order, count(0, 1)))

        self._shape = (len(self.tracks), len(self.samples))

        if custom_pieces is None:
            custom_pieces = {}
        self.custom_pieces = custom_pieces

        self._process_alterations()

        layers = {}
        for e in self.events:
            layers[e] = np.zeros(self._shape, dtype=bool)

        for _, row in self.data.iterrows():
            patient, track, event = row
            row_ix = self._tracks_ix[track]
            col_ix = self._patients_ix[patient]
            layers[event][row_ix, col_ix] = True

        self.layers = layers

    def _process_alterations(self):
        events_alt = dict()
        custom_events = list(self.custom_pieces.keys())
        raw_events = self.data["event"].unique()

        unknown_alterations = []
        for e in raw_events:
            alt = guess_alteration(e)
            if alt == Alteration.OTHER:
                alt = e
                if e not in custom_events:
                    unknown_alterations.append(e)
            events_alt[e] = alt
        self.data["event"] = [events_alt[e] for e in self.data["event"]]
        self.events = self.data["event"].unique()
        if len(unknown_alterations) > 0:
            msg = (
                f"Found unknown alterations: {unknown_alterations}, "
                f"please specify a piece for this alteration."
            )
            warnings.warn(msg)

    @property
    def shape(self):
        return self._shape

    def get_layers_data(self, background_color="#BEBEBE"):
        # explicitly make copy
        bg_pieces = deepcopy(SHAPE_BANK[Alteration.BACKGROUND])
        bg_pieces.background_color = background_color

        # Add background layer
        layers_data = [
            LayerData(
                np.ones(self._shape), bg_pieces, background_color, Alteration.BACKGROUND
            )
        ]
        for alt in self.events:
            layer = self.layers[alt]
            if not isinstance(alt, Alteration):
                piece = self.custom_pieces.get(alt)
                if piece is None:
                    # The default style for OTHER
                    piece = FrameRect(color="pink", label=alt)
                else:
                    piece = deepcopy(piece)
            else:
                piece = deepcopy(SHAPE_BANK[alt])
            color = piece.color
            piece.background_color = background_color

            layers_data.append(LayerData(layer, piece, color, alt))

        return layers_data

    def get_track_mutation_rate(self):
        gb = self.data.groupby("track", sort=False, observed=True)
        ts = {}
        for track, df in gb:
            ts[track] = len(df["sample"].unique())
        counts = np.array([ts[t] for t in self.tracks])
        return counts / len(self.samples)

    def get_track_mutation_types(self):
        gb = self.data.groupby("track", sort=False, observed=True)
        cs = {}
        for track, df in gb:
            cs[track] = Counter(df["event"])
        return pd.DataFrame(cs).fillna(0.0)

    def get_sample_mutation_types(self):
        gb = self.data.groupby("sample", sort=False, observed=True)
        cs = {}
        for track, df in gb:
            cs[track] = Counter(df["event"])
        return pd.DataFrame(cs).fillna(0.0)


class OncoPrint(ClusterBoard):
    """OncoPrint plot

    The oncoprint plot is a visualization for genomics data in cancer research.
    It's first introduced by the cBioPortal project.
    See https://www.cbioportal.org/oncoprinter for more details.

    To use this class, import from oncoprinter

        >>> from oncoprinter import OncoPrint

    Parameters
    ----------
    genomic_data : pd.DataFrame
        Genomics data, each column is:
            1) Sample ID
            2) Track name
            3) Alteration

    patients_order : list, optional
        The order of samples, by default None
    tracks_order : list, optional
        The order of tracks, by default None
    pieces : dict, optional
        Custom pieces for each alteration, by default None
        See :class:`Piece <marsilea.layers.Piece>` for details
    background_color : str, optional, default: "#BEBEBE"
        The background color
    shrink : tuple, optional, default: (0.8, 0.8)
        The shrink ratio for each layer
    width, height : float, optional
        The size in inches to define the size of main canvas
    aspect : float, optional, default: 2.5
        The aspect ratio of the main canvas
    legend_kws : dict, optional
        The options for legend, by default None
        See :class:`cat_legend <legendkit.cat_legend>` for details
    name : str, optional
        The name of this OncoPrint
    add_tracks_names : str, optional, default: "left"
        The position to add tracks names
        If None, will not add tracks names
    add_samples_names : str, optional, default: "bottom"
        The position to add samples names
        If None, will not add samples names
    add_mut_perc : str, optional, default: "right"
        The position to add mutation percentage
        If None, will not add mutation percentage
    add_tracks_counts : str, optional, default: "right"
        The position to add tracks mutation counts
        If None, will not add tracks mutation counts
    add_mut_counts : str, optional, default: "top"
        The position to add mutation counts
        If None, will not add mutation counts
    add_tracks_counts_size : float, optional, default: 0.2
        The size of tracks mutation counts
    add_tracks_counts_pad : float, optional, default: 0
        The padding of tracks mutation counts
    add_mut_counts_size : float, optional, default: 0.2
        The size of mutation counts
    add_mut_counts_pad : float, optional, default: 0.1
        The padding of mutation counts

    """

    def __init__(
        self,
        genomic_data=None,
        patients_order=None,
        tracks_order=None,
        pieces=None,
        background_color="#BEBEBE",
        shrink=(0.8, 0.8),
        width=None,
        height=None,
        aspect=2.5,
        legend_kws=None,
        name=None,
        add_tracks_names="left",
        add_samples_names="bottom",
        add_mut_perc="right",
        add_tracks_counts="right",
        add_mut_counts="top",
        add_tracks_counts_size=0.2,
        add_tracks_counts_pad=0,
        add_mut_counts_size=0.2,
        add_mut_counts_pad=0.1,
    ):
        data = GenomicData(
            genomic_data,
            samples_order=patients_order,
            tracks_order=tracks_order,
            custom_pieces=pieces,
        )
        self.genomic_data = data
        width, height = get_canvas_size_by_data(
            data.shape, width=width, height=height, scale=0.2, aspect=aspect
        )

        self.canvas = super().__init__(
            name=name, cluster_data=np.zeros(data.shape), width=width, height=height
        )

        legend_options = dict(title="Alterations", handleheight=aspect, handlelength=1)
        legend_kws = {} if legend_kws is None else legend_kws
        legend_options.update(legend_kws)

        layers, pieces, colors_mapper = [], [], {}
        for layer in data.get_layers_data(background_color):
            layers.append(layer.matrix)
            pieces.append(layer.piece)
            colors_mapper[layer.event] = layer.color
        mesh = LayersMesh(
            layers=layers, pieces=pieces, shrink=shrink, legend_kws=legend_options
        )
        self.add_layer(mesh)

        if add_tracks_names:
            self.add_plot(add_tracks_names, Labels(data.tracks))

        # Add other statistics
        track_mut_rate = data.get_track_mutation_rate()
        # Convert to percentage string
        if add_mut_perc:
            rates = [_format_percentage(t) for t in track_mut_rate]
            self.add_plot(add_mut_perc, Labels(rates))
        if add_samples_names:
            self.add_plot(add_samples_names, Labels(data.samples))

        if add_tracks_counts:
            track_counter = data.get_track_mutation_types()
            colors = [colors_mapper[e] for e in track_counter.index]
            track_bar = StackBar(track_counter, colors=colors, show_value=False)
            self.add_plot(
                add_tracks_counts,
                track_bar,
                legend=False,
                size=add_tracks_counts_size,
                pad=add_tracks_counts_pad,
            )

        if add_mut_counts:
            patients_counter = data.get_sample_mutation_types()
            colors = [colors_mapper[e] for e in patients_counter.index]
            patients_bar = StackBar(patients_counter, colors=colors, show_value=False)
            self.add_plot(
                add_mut_counts,
                patients_bar,
                legend=False,
                size=add_mut_counts_size,
                pad=add_mut_counts_pad,
            )
        self.add_legends()

    clinical_plots = {
        "bar": Numbers,
        "stack_bar": StackBar,
    }

    def add_clinical_data(self, data, plot="bar", size=None, pad=0.1, **kwargs):
        data = data.loc[self.samples_order]
        plotter = self.clinical_plots[plot]
        self.add_bottom(plotter(data, **kwargs), size=size, pad=pad)

    def add_heatmap_data(self, data, size=0.2, pad=0.1, **kwargs):
        data = data.loc[self.samples_order]
        options = {"cmap": "Blues", "label_loc": "left", **kwargs}
        self.add_bottom(ColorMesh(data, **options), size=size, pad=pad)

    @property
    def samples_order(self):
        return self.genomic_data.samples

    @property
    def tracks_order(self):
        return self.genomic_data.tracks


def _format_percentage(t):
    return f"{float(t) * 100:.2f}".rstrip("0").rstrip(".") + "%"



__all__ = ["ColorMesh", "Colors", "SizedMesh", "MarkerMesh", "TextMesh", "PatchMesh"]

import numpy as np
import pandas as pd
import warnings
from itertools import cycle
from legendkit import ColorArt, CatLegend, SizeLegend
from matplotlib.cm import ScalarMappable
from matplotlib.colors import ListedColormap, TwoSlopeNorm, Normalize, is_color_like
from typing import Mapping

from ._utils import _format_label
from .base import RenderPlan
from ..layout import close_ticks
from ..utils import relative_luminance, get_colormap, ECHARTS16, get_canvas_size_by_data


def _mask_data(data, mask=None):
    if isinstance(data, pd.DataFrame):
        data = data.to_numpy()
    else:
        data = np.asarray(data)
    if mask is not None:
        data = np.ma.masked_where(np.asarray(mask), data)
    if data.ndim == 1:
        data = data.reshape(1, -1)
    return data


class MeshBase(RenderPlan):
    norm = None
    cmap = None
    render_main = True
    _legend_kws = {}

    def _process_cmap(self, data, vmin, vmax, cmap, norm, center):
        self.cmap = "coolwarm" if cmap is None else cmap
        self.norm = norm

        self.vmin = np.nanmin(data) if vmin is None else vmin
        self.vmax = np.nanmax(data) if vmax is None else vmax

        if norm is None:
            self.norm = Normalize(vmin=self.vmin, vmax=self.vmax)

        if center is not None:
            self.norm = TwoSlopeNorm(center, vmin=vmin, vmax=vmax)

    def set_legends(self, **kwargs):
        self._legend_kws.update(kwargs)


class ColorMesh(MeshBase):
    """Continuous Heatmap

    Parameters
    ----------
    data : np.ndarray, pd.DataFrame
        2D data
    cmap : str or :class:`matplotlib.colors.Colormap`
        The colormap used to map the value to the color
    norm : :class:`matplotlib.colors.Normalize`
        A Normalize instance to map data
    vmin, vmax : number, optional
        Value to determine the value mapping behavior
    center : number, optional
        The value to center on colormap, notice that this is different from
        seaborn; If set, a :class:`matplotlib.colors.TwoSlopeNorm` will be used
        to center the colormap.
    mask : array-like
        A bool matrix indicates which cell will be masked and not rendered
    alpha : float
        The transparency value
    linewidth : float
        The width of grid line
    linecolor : color
        The color of grid line
    annot : bool, array-like
        Whether to show the value / The text to show in each cell
    fmt : str, callable
        The format string or a function to format the value
    annot_kws : dict
        See :class:`matplotlib.text.Text`
    cbar_kws : dict
        See :class:`matplotlib.colorbar.Colorbar`,
    label : str
        The label of the plot, only show if added to the side plot
    label_loc : {'top', 'bottom', 'left', 'right'}
        Where to add the label
    label_props : dict
        See :class:`matplotlib.text.Text`
    kwargs :

    See Also
    --------
        :class:`marsilea.Heatmap`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> from marsilea.plotter import ColorMesh
        >>> _, ax = plt.subplots(figsize=(5, 0.5))
        >>> ColorMesh(np.arange(10), cmap="Blues").render(ax)

    .. plot::
        :context: close-figs

        >>> data = np.random.randn(10, 8)
        >>> h = ma.Heatmap(data)
        >>> h.cut_rows(cut=[5])
        >>> h.add_dendrogram("left")
        >>> cmap1, cmap2 = "Purples", "Greens"
        >>> colors1 = ColorMesh(np.arange(10) + 1, cmap=cmap1, label=cmap1, annot=True)
        >>> colors2 = ColorMesh(np.arange(10) + 1, cmap=cmap2, label=cmap2)
        >>> h.add_right(colors1, size=0.2, pad=0.05)
        >>> h.add_right(colors2, size=0.2, pad=0.05)
        >>> h.render()


    """

    def __init__(
        self,
        data,
        cmap=None,
        norm=None,
        vmin=None,
        vmax=None,
        mask=None,
        center=None,
        alpha=None,
        linewidth=None,
        linecolor=None,
        annot=None,
        fmt=None,
        annot_kws=None,
        cbar_kws=None,
        label=None,
        label_loc=None,
        label_props=None,
        **kwargs,
    ):
        self.data = _mask_data(data, mask)
        self.alpha = alpha
        self.linewidth = linewidth
        self.linecolor = linecolor
        self.annotated_texts = self.data
        self.annot = False
        if annot is not None:
            if isinstance(annot, bool):
                self.annot = annot
            else:
                self.annot = True
                self.annotated_texts = annot
        self.fmt = ".2g" if fmt is None else fmt
        self.annot_kws = {} if annot_kws is None else annot_kws
        self._process_cmap(data, vmin, vmax, cmap, norm, center)

        self.set_label(label, label_loc, label_props)
        self.set_data(self.data, self.annotated_texts)
        self._legend_kws = dict(title=self.label)
        if cbar_kws is not None:
            self.set_legends(**cbar_kws)

        self.kwargs = kwargs

    def _annotate_text(self, ax, mesh, texts):
        """Add textual labels with the value in each cell."""
        mesh.update_scalarmappable()
        height, width = texts.shape
        xpos, ypos = np.meshgrid(np.arange(width) + 0.5, np.arange(height) + 0.5)
        for x, y, m, color, val in zip(
            xpos.flat,
            ypos.flat,
            mesh.get_array().flatten(),
            mesh.get_facecolors(),
            texts.flat,
        ):
            if m is not np.ma.masked:
                lum = relative_luminance(color)
                text_color = ".15" if lum > 0.408 else "w"
                annotation = _format_label(val, self.fmt)
                text_kwargs = dict(color=text_color, ha="center", va="center")
                text_kwargs.update(self.annot_kws)
                ax.text(x, y, annotation, **text_kwargs)

    def get_legends(self):
        mappable = ScalarMappable(norm=self.norm, cmap=self.cmap)
        return ColorArt(mappable, **self._legend_kws)

    def render_ax(self, spec):
        values, texts = spec.data
        ax = spec.ax
        if self.is_flank:
            values = values.T
            texts = texts.T
        mesh = ax.pcolormesh(
            values,
            cmap=self.cmap,
            norm=self.norm,
            linewidth=self.linewidth,
            edgecolor=self.linecolor,
            **self.kwargs,
        )
        if self.annot:
            self._annotate_text(ax, mesh, texts)
        # set the mesh for legend

        if not ax.yaxis_inverted():
            ax.invert_yaxis()
        ax.set_axis_off()


# transform input data to numeric
def encode_numeric(arr, encoder):
    orig_shape = arr.shape
    if np.ma.isMaskedArray(arr):
        re_arr = []
        for e in arr.flatten():
            if e is np.ma.masked:
                re_arr.append(np.nan)
            else:
                re_arr.append(encoder[e])
        re_arr = np.ma.masked_invalid(re_arr)
    else:
        re_arr = [encoder[e] for e in arr.flatten()]
    re_arr = np.array(re_arr).reshape(orig_shape)
    return re_arr


def _enough_colors(n_colors, n_cats):
    # Inform the user if the same color
    # will be used more than once
    if n_colors < n_cats:
        warnings.warn(
            f"Current colormap has only {n_colors} colors "
            f"which is less that your input "
            f"with {n_cats} elements"
        )


class Colors(MeshBase):
    """Categorical colors/heatmap

    Parameters
    ----------
    data : np.ndarray
    label : str
    label_loc : {'top', 'bottom', 'left', 'right'}
    palette : dict, array-like
        Could be a mapping of label and colors or an array match
        the shape of data
    cmap : str or :class:`matplotlib.colors.Colormap`
        The colormap used to map the value to the color
    mask : array of bool
        Indicate which cell will be masked and not rendered
    label : str
        The label of the plot, only show if added to the side plot
    label_loc : str
        Where to add the label
    label_props : dict
        See :class:`matplotlib.text.Text`
    legend_kws :
        See :class:`legendkit.legend`
    kwargs :
        Pass to :meth:`pcolormesh <matplotlib.axes.Axes.pcolormesh>`

    See Also
    --------
        :class:`marsilea.CatHeatmap`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> from marsilea.plotter import Colors
        >>> _, ax = plt.subplots(figsize=(5, 0.5))
        >>> data = np.random.choice(["A", "B", "C"], 10)
        >>> Colors(data).render(ax)

    .. plot::
        :context: close-figs

        >>> h = ma.Heatmap(np.random.randn(10, 8))
        >>> h.cut_rows(cut=[5])
        >>> h.add_dendrogram("left")
        >>> color = Colors(data, label="Colors")
        >>> h.add_right(color, size=0.2, pad=0.05)
        >>> h.render()


    """

    def __init__(
        self,
        data,
        palette=None,
        cmap=None,
        mask=None,
        linewidth=None,
        linecolor=None,
        label=None,
        label_loc=None,
        label_props=None,
        legend_kws=None,
        **kwargs,
    ):
        data = np.asarray(data)
        self.set_label(label, label_loc, label_props)
        self.linewidth = linewidth
        self.linecolor = linecolor
        self.kwargs = kwargs

        self._legend_kws = dict(title=self.label, size=1)
        if legend_kws is not None:
            self.set_legends(**legend_kws)

        encoder = {}
        render_colors = []
        # get unique color
        if palette is not None:
            if isinstance(palette, Mapping):
                self.palette = palette
            else:
                palette = np.asarray(palette)
                self.palette = dict(zip(np.unique(data), palette.flat))

            for i, (label, color) in enumerate(self.palette.items()):
                encoder[label] = i
                render_colors.append(color)

        else:
            cats = np.unique(data)
            if cmap is not None:
                cmap = get_colormap(cmap)
                colors = cmap(np.linspace(0, 1, len(cats)))
            else:
                colors = []
                for i, c in enumerate(cycle(ECHARTS16)):
                    colors.append(c)
                    if i == len(cats):
                        break

            _enough_colors(len(colors), len(cats))
            palette = {}
            render_colors = []
            for i, (label, color) in enumerate(zip(cats, cycle(colors))):
                encoder[label] = i
                render_colors.append(color)
                palette[label] = color
            self.palette = palette

        self.render_cmap = ListedColormap(render_colors)
        self.vmax = len(render_colors) - 1
        # Encode data into cmap range
        encode_data = encode_numeric(data, encoder)
        self.set_data(_mask_data(encode_data, mask=mask))
        # If the data is numeric, we don't change it
        if np.issubdtype(data.dtype, np.number):
            self.cluster_data = data
        else:
            self.cluster_data = encode_data

    def get_legends(self):
        colors = []
        labels = []
        for label, color in self.palette.items():
            labels.append(label)
            colors.append(color)
        return CatLegend(colors=colors, labels=labels, **self._legend_kws)

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        if self.is_flank:
            data = data.T
        ax.pcolormesh(
            data,
            cmap=self.render_cmap,
            linewidth=self.linewidth,
            edgecolor=self.linecolor,
            vmin=0,
            vmax=self.vmax,
            **self.kwargs,
        )
        ax.set_axis_off()
        if not ax.yaxis_inverted():
            ax.invert_yaxis()


class SizedMesh(MeshBase):
    """Mesh for sized elements

    .. note::
        If encode color as categorical data, `palette` must be used
        to assign color for each category.

    Parameters
    ----------
    size : 2d array
        Control the radius of circles, must be numeric
    color: color-like or 2d array
        The color of circles, could be numeric / categorical matrix
        If using one color name, all circles will have the same color.
    cmap : str or :class:`matplotlib.colors.Colormap`
        The colormap used to map the value to the color
    norm : :class:`matplotlib.colors.Normalize`
        A Normalize instance to map data
    vmin, vmax : number, optional
        Value to determine the value mapping behavior
    center : number, optional
        The value to center on colormap, notice that this is different from
        seaborn; If set, a :class:`matplotlib.colors.TwoSlopeNorm` will be used
        to center the colormap.
    alpha : float
        The transparency value
    sizes : tuple of float
        The range of the size of elements
    size_norm : :class:`matplotlib.colors.Normalize`
        A Normalize instance to map size
    edgecolor : color
        The border color of each elements
    linewidth : float
        The width of the border of each elements
    frameon : bool
        Whether to draw a frame
    grid : bool
        Whether to draw grid
    grid_color : color
        The color of grid
    grid_linewidth : float
        The width of grid line
    palette : dict
        Use to map color if input categorical data
    marker : str
        See :mod:`matplotlib.markers`
    label : str
        The label of the plot, only show if added to the side plot
    label_loc : {'top', 'bottom', 'left', 'right'}
        Where to add the label
    label_props : dict
        See :class:`matplotlib.text.Text`
    legend : bool, default: True
        Whether to add a legend
    size_legend_kws : dict
        Control the size legend, See :class:`legendkit.size_legend`
    color_legend_kws : dict
        Control the color legend, See :class:`legendkit.colorart`
    kwargs : dict
        Pass to :meth:`matplotlib.axes.Axes.scatter`

    See Also
    --------
        :class:`marsilea.SizedHeatmap`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> from marsilea.plotter import SizedMesh
        >>> _, ax = plt.subplots(figsize=(5, 0.5))
        >>> size, color = np.random.rand(1, 10), np.random.rand(1, 10)
        >>> SizedMesh(size, color).render(ax)

    .. plot::
        :context: close-figs

        >>> h = ma.Heatmap(np.random.randn(10, 8))
        >>> h.cut_rows(cut=[5])
        >>> h.add_dendrogram("left")
        >>> mesh = SizedMesh(size, color, marker="*", label="SizedMesh")
        >>> h.add_right(mesh, size=0.2, pad=0.05)
        >>> h.render()



    """

    def __init__(
        self,
        size,
        color=None,
        cmap=None,
        norm=None,
        vmin=None,
        vmax=None,
        alpha=None,
        center=None,
        sizes=(1, 200),
        size_norm=None,
        edgecolor=None,
        linewidth=1,
        frameon=True,
        grid=False,
        grid_color=".8",
        grid_linewidth=1,
        palette=None,
        marker="o",
        label=None,
        label_loc=None,
        label_props=None,
        legend=True,
        size_legend_kws=None,
        color_legend_kws=None,
        **kwargs,
    ):
        # normalize size
        size = np.asarray(size)
        if size_norm is None:
            size_norm = Normalize()
            size_norm.autoscale(size)
        self.size_norm = size_norm
        self.orig_size = size
        self.size_matrix = size_norm(size) * (sizes[1] - sizes[0]) + sizes[0]
        self.color = None
        self.color2d = None
        self.palette = palette
        self.marker = marker
        self.legend = legend
        self.color_legend_kws = {} if color_legend_kws is None else color_legend_kws
        self.size_legend_kws = {} if size_legend_kws is None else size_legend_kws
        self._has_colormesh = False
        # process color
        # By default, the circles colors are uniform
        if color is None:
            color = "C0"
        if color is not None:
            if is_color_like(color):
                self.color = color
                self.color2d = np.repeat(color, size.size).reshape(size.shape)
            else:
                color = np.asarray(color)
                # categorical circle color
                if palette is not None:
                    encoder = {}
                    render_colors = []
                    for i, (plabel, c) in enumerate(palette.items()):
                        encoder[plabel] = i
                        render_colors.append(c)
                    self.cmap = ListedColormap(render_colors)
                    self.color2d = encode_numeric(color, encoder)
                else:
                    self._process_cmap(color, vmin, vmax, cmap, norm, center)
                    self.color2d = color
                self._has_colormesh = True
        self.alpha = alpha
        self.frameon = frameon
        self.edgecolor = edgecolor
        self.linewidth = linewidth
        self.set_label(label, label_loc, label_props)
        self.grid = grid
        self.grid_color = grid_color
        self.grid_linewidth = grid_linewidth
        self.kwargs = kwargs

        self._collections = None
        self.set_data(self.size_matrix, self.color2d)

    def update_main_canvas_size(self):
        return get_canvas_size_by_data(self.orig_size.shape)

    def get_legends(self):
        if not self.legend:
            return None
        if self.color is not None:
            size_color = self.color
        else:
            size_color = "black"
        handler_kw = dict(edgecolor=self.edgecolor, linewidth=self.linewidth)
        options = dict(
            colors=size_color,
            handle=self.marker,
            show_at=(0.1, 0.25, 0.5, 0.75, 1.0),
            spacing="uniform",
            handler_kw=handler_kw,
        )
        options.update(self.size_legend_kws)
        size_legend = SizeLegend(self.size_matrix, array=self.orig_size, **options)

        if self._has_colormesh & (self.color != "none"):
            if self.palette is not None:
                labels, colors = [], []
                for label, c in self.palette.items():
                    labels.append(label)
                    colors.append(c)
                options = dict(
                    size=1,
                    handle=self.marker,
                    handler_kw=handler_kw,
                    frameon=False,
                )
                options.update(self.color_legend_kws)
                color_legend = CatLegend(labels=labels, colors=colors, **options)
            else:
                ScalarMappable(norm=self.norm, cmap=self.cmap)
                color_legend = ColorArt(self._collections, **self.color_legend_kws)
            return [size_legend, color_legend]
        else:
            return size_legend

    def render_ax(self, spec):
        ax = spec.ax
        size, color = spec.data
        if self.is_flank:
            size = size.T
            color = color.T
        Y, X = size.shape
        xticks = np.arange(X) + 0.5
        yticks = np.arange(Y) + 0.5
        xv, yv = np.meshgrid(xticks, yticks)

        if self.grid:
            for x in xticks:
                ax.axvline(
                    x, color=self.grid_color, linewidth=self.grid_linewidth, zorder=0
                )
            for y in yticks:
                ax.axhline(
                    y, color=self.grid_color, linewidth=self.grid_linewidth, zorder=0
                )

        options = dict(
            s=size,
            edgecolor=self.edgecolor,
            linewidths=self.linewidth,
            marker=self.marker,
        )

        if self.color is not None:
            options["c"] = self.color
        else:
            options.update(dict(c=color, norm=self.norm, cmap=self.cmap))
        self._collections = ax.scatter(xv, yv, **options, **self.kwargs)

        close_ticks(ax)
        if not self.frameon:
            ax.set_axis_off()
        ax.set_xlim(0, xticks[-1] + 0.5)
        ax.set_ylim(0, yticks[-1] + 0.5)
        if not ax.yaxis_inverted():
            ax.invert_yaxis()


# TODO: A patch mesh
class PatchMesh(MeshBase):
    pass


class MarkerMesh(MeshBase):
    """The mesh that draw marker shape

    Parameters
    ----------

    data : np.ndarray
        Must be bool matrix to indicate if a marker is drawn at specific cell
    color : color
        The color of the marker.
    marker : str
        See :mod:`matplotlib.markers`
    size : int
        The of marker in fontsize unit
    frameon : bool
        Whether to draw the border of the plot
    label : str
        The label of the plot, only show when added to the side plot
    label_loc : {'top', 'bottom', 'left', 'right'}
        The position of the label
    label_props : dict
        See :class:`matplotlib.text.Text`
    kwargs :


    Examples
    --------

    .. plot::
        :context: close-figs

        >>> from marsilea.plotter import MarkerMesh
        >>> data = np.random.randn(10, 10) > 0
        >>> _, ax = plt.subplots(figsize=(3, 3))
        >>> MarkerMesh(data, color="darkgreen", marker="x", size=50).render(ax)

    """

    render_main = True

    def __init__(
        self,
        data,
        color="black",
        marker="*",
        size=35,
        frameon=False,
        label=None,
        label_loc=None,
        label_props=None,
        **kwargs,
    ):
        self.set_data(np.asarray(data))
        self.color = color
        self.marker = marker
        self.marker_size = size
        self.frameon = frameon
        self.set_label(label, label_loc, label_props)
        self.kwargs = kwargs

    def get_legends(self):
        return CatLegend(
            colors=[self.color], labels=[self.label], handle=self.marker, draw=False
        )

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        Y, X = data.shape
        xticks = np.arange(X) + 0.5
        yticks = np.arange(Y) + 0.5
        yv, xv = np.where(data)

        ax.scatter(
            xv + 0.5,
            yv + 0.5,
            s=self.marker_size,
            c=self.color,
            marker=self.marker,
            **self.kwargs,
        )

        close_ticks(ax)
        ax.set_xlim(0, xticks[-1] + 0.5)
        ax.set_ylim(0, yticks[-1] + 0.5)
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
        if not self.frameon:
            ax.set_axis_off()


class TextMesh(MeshBase):
    """The mesh that draw text on each cell

    Parameters
    ----------
    texts : np.ndarray
        The text to draw
    color : color
        The color of the text
    frameon : bool
        Whether to draw the border of the plot
    label : str
        The label of the plot, only show when added to the side plot
    label_loc : {'top', 'bottom', 'left', 'right'}
        The position of the label
    label_props : dict
        Props for label :class:`matplotlib.text.Text`
    kwargs : dict
        Pass to :meth:`matplotlib.axes.Axes.text`

    """

    render_main = True

    def __init__(
        self,
        texts,
        color="black",
        frameon=False,
        label=None,
        label_loc=None,
        label_props=None,
        **kwargs,
    ):
        self.set_data(self.data_validator(texts))
        self.color = color
        self.frameon = frameon
        self.set_label(label, label_loc, label_props)
        self.kwargs = kwargs

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        Y, X = data.shape
        xticks = np.arange(X) + 0.5
        yticks = np.arange(Y) + 0.5

        text_options = {
            "ha": "center",
            "va": "center",
            "color": self.color,
            **self.kwargs,
        }
        for x in range(X):
            for y in range(Y):
                ax.text(x + 0.5, y + 0.5, data[y, x], **text_options)

        close_ticks(ax)
        ax.set_xlim(0, xticks[-1] + 0.5)
        ax.set_ylim(0, yticks[-1] + 0.5)
        if not ax.yaxis_inverted():
            ax.invert_yaxis()
        if not self.frameon:
            ax.set_axis_off()



import numpy as np


def _format_labels(labels, fmt):
    f_labels = []
    for a in labels:
        label = _format_label(a, fmt)
        f_labels.append(label)
    return f_labels


def _format_label(a, fmt):
    if np.isnan(a):
        a = ""
    if isinstance(fmt, str):
        label = _auto_format_str(fmt, a)
    elif callable(fmt):
        label = fmt(a)
    else:
        raise TypeError("fmt must be a str or callable")
    return label


def _auto_format_str(fmt, value):
    """Format a value according to the given format string.
    matplotlib/cbook.py
    """
    try:
        fmt = "{:" + fmt + "}"
        return fmt.format(value)
    except (TypeError, ValueError):
        return str(value)



class DuplicateName(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"Axes with name `{self.name}` already exists."


class DuplicatePlotter(Exception):
    def __init__(self, plotter):
        self.plotter = plotter

    def __str__(self):
        name = self.plotter.__class__.__name__
        return (
            f"The `{name}` added"
            f"to the `{self.plotter.side}` has been registered,"
            f"please create a new `{name}` if you want to add the same plotter again."
        )


class SplitTwice(Exception):
    def __init__(self, axis="col"):
        self.axis = axis

    def __str__(self):
        return f"Split {self.axis} more than once is ambiguous."


class SplitConflict(Exception):
    pass


class AppendLayoutError(Exception):
    def __str__(self):
        return (
            "Append a concatenated plot is not allowed,"
            "you can only append "
            "plots to a concatenated plot."
        )


class DataError(Exception):
    pass



import numpy as np
from itertools import cycle
from matplotlib.collections import LineCollection
from matplotlib.colors import is_color_like
from matplotlib.lines import Line2D
from scipy.cluster.hierarchy import linkage as scipy_linkage, dendrogram
from typing import List, Sequence


class _DendrogramBase:
    is_singleton = False

    def __init__(
        self,
        data,
        method=None,
        metric=None,
        linkage=None,
        get_meta_center=None,
        key=None,
        **kwargs,
    ):
        self.key = key
        self.data = data
        if method is None:
            method = "single"
        if metric is None:
            metric = "euclidean"
        # edge case: data is 1d, may happen when user split the data
        if len(data) == 1:
            # TODO: the y coords are heuristic value,
            #       need a better way to handle
            self.x_coords = np.array([[1.0, 1.0, 1.0, 1.0]])
            self.y_coords = np.array([[0.0, 0.75, 0.75, 0.0]])
            self._reorder_index = np.array([0])
            self.is_singleton = True
        else:
            if linkage is not None:
                self.Z = linkage
            else:
                self.Z = scipy_linkage(data, method=method, metric=metric)
            self._plot_data = dendrogram(self.Z, no_plot=True)

            self.x_coords = np.asarray(self._plot_data["icoord"]) / 5
            self.y_coords = np.asarray(self._plot_data["dcoord"])
            self._reorder_index = self._plot_data["leaves"]

            if len(self.y_coords) == 1:
                self.y_coords = np.array([[0.0, 0.75, 0.75, 0.0]])
            else:
                ycoords = np.unique(self.y_coords)
                ycoords = ycoords[np.nonzero(ycoords)]
                y_min, y_max = np.min(ycoords), np.max(ycoords)
                interval = y_max - y_min
                for i, j in zip(*np.nonzero(self.y_coords)):
                    if self.y_coords[i, j] != 0.0:
                        v = self.y_coords[i, j]
                        self.y_coords[i, j] = (v - y_min) / interval + 0.2
                    # self.y_coords[i, j] -= offset
            # self.y_coords[np.nonzero(self.y_coords)] - offset

        self._render_x_coords = self.x_coords
        self._render_y_coords = self.y_coords
        self.n_leaves = len(self.reorder_index)
        self.max_dependent_coord = np.max(self.y_coords)

        self.xlim = np.array([0, self.n_leaves * 2])
        self.ylim = np.array([0, self.max_dependent_coord * 1.05])
        self._render_xlim = self.xlim
        self._render_ylim = self.ylim

        # Should be lazy eval
        # TODO: Allow center to be calculated differently
        if get_meta_center is None:
            self._center = np.mean(data, axis=0)
        elif callable(get_meta_center):
            # Ensure the centroid function returns a numpy array of correct shape
            centroid = get_meta_center(data)
            if isinstance(centroid, np.ndarray) and centroid.shape == data.shape[1:]:
                self._center = centroid
            else:
                raise ValueError(
                    "The get_meta_center must return a numpy array with shape "
                    "matching the number of features in the data."
                )
        else:
            raise TypeError("The get_meta_center must be a callable function or None.")

    @property
    def xrange(self):
        return self.xlim[1] - self.xlim[0]

    @property
    def yrange(self):
        return self.ylim[1] - self.ylim[0]

    @property
    def render_xrange(self):
        return self._render_xlim[1] - self._render_xlim[0]

    @property
    def render_yrange(self):
        return self._render_ylim[1] - self._render_ylim[0]

    def set_lim(self, x_start=None, y_end=None):
        if x_start is not None:
            self._render_xlim = self.xlim + x_start
            self._render_x_coords = self.x_coords + x_start

        if y_end is not None:
            if y_end < self.ylim[1]:
                raise ValueError(
                    f"{y_end} is lower than " f"current ylim at {self.ylim[1]}"
                )
            self._render_ylim = (0, y_end)

    @property
    def reorder_index(self):
        return self._reorder_index

    @property
    def center(self):
        return self._center

    @property
    def root(self):
        xc = self.x_coords[-1]
        yc = self.y_coords[-1]
        x1 = (xc[2] - xc[1]) / 2.0 + xc[1]
        y1 = yc[1]
        return x1, y1

    @property
    def render_root(self):
        xc = self._render_x_coords[-1]
        yc = self._render_y_coords[-1]
        x1 = (xc[2] - xc[1]) / 2.0 + xc[1]
        y1 = yc[1]
        return x1, y1

    def _draw_dendrogram(
        self, ax, orient="top", color=".1", linewidth=0.7, rasterized=False
    ):
        x_coords = self._render_x_coords
        y_coords = self._render_y_coords
        if orient in ["right", "left"]:
            x_coords, y_coords = y_coords, x_coords

        lines = LineCollection(
            [list(zip(x, y)) for x, y in zip(x_coords, y_coords)],
            color=color,
            linewidth=linewidth,
            rasterized=rasterized,
        )
        ax.add_collection(lines)


class Dendrogram(_DendrogramBase):
    """A dendrogram class

    Parameters
    ----------

    data : np.ndarray
    method : str
        Refer to :func:`scipy.cluster.hierarchy.linkage`
    metric : str
        Refer to :func:`scipy.cluster.hierarchy.linkage`

    """

    def __init__(
        self,
        data: np.ndarray,
        method=None,
        metric=None,
        linkage=None,
        get_meta_center=None,
        key=None,
        **kwargs,
    ):
        super().__init__(
            data,
            method=method,
            metric=metric,
            key=key,
            linkage=linkage,
            get_meta_center=get_meta_center,
            **kwargs,
        )

    # here we left an empty **kwargs to align api with GroupDendrogram
    def draw(
        self,
        ax,
        orient="top",
        color=None,
        linewidth=None,
        add_root=False,
        root_color=None,
        control_ax=True,
        rasterized=False,
        **kwargs,
    ):
        """

        Parameters
        ----------
        ax
        orient
        color : color
            The line color of the dendrogram
        linewidth : float
        add_root : bool
            Add a line to represent the root of dendrogram
        root_color : color
            The color of the root line
        control_ax : bool
            Adjust the axes to ensure the dendrogram will display correctly
        rasterized : bool
            Rasterize the dendrogram to speed up rendering

        Returns
        -------

        """
        color = ".1" if color is None else color
        root_color = color if root_color is None else root_color
        linewidth = 0.7 if linewidth is None else linewidth

        self._draw_dendrogram(
            ax, orient=orient, color=color, linewidth=linewidth, rasterized=rasterized
        )

        xlim = self._render_xlim
        ylim = self._render_ylim
        if orient in ["right", "left"]:
            xlim, ylim = ylim, xlim

        if control_ax:
            ax.set_xlim(*xlim)
            ax.set_ylim(*ylim)
            if orient == "left":
                ax.invert_xaxis()
            if orient == "bottom":
                ax.invert_yaxis()

        if add_root:
            x1, y1 = self.render_root
            if orient in ["right", "left"]:
                x1, y1 = y1, x1
                x2 = xlim[1]
                y2 = y1
            else:
                x2 = x1
                y2 = ylim[1]
            root_line = Line2D(
                [x1, x2],
                [y1, y2],
                color=root_color,
                linewidth=linewidth,
                rasterized=rasterized,
            )
            ax.add_artist(root_line)


class GroupDendrogram(_DendrogramBase):
    """Meta dendrogram

    Parameters
    ----------

    dens : array of :class:`Dendrogram`
        A list of :class:`Dendrogram`
    method : str
    metric : str

    """

    def __init__(
        self,
        dens: List[Dendrogram],
        method=None,
        metric=None,
        linkage=None,
        get_meta_center=None,
        key=None,
        **kwargs,
    ):
        data = np.vstack([d.center for d in dens])
        super().__init__(
            data,
            method=method,
            metric=metric,
            linkage=linkage,
            get_meta_center=get_meta_center,
            key=key,
            **kwargs,
        )
        self.orig_dens = np.asarray(dens)
        self.dens = np.asarray(dens)[self.reorder_index]
        self.n = len(self.dens)

        den_xlim = 0
        ylim = 0
        x_coords = []
        for den in self.dens:
            den_xlim += den.xrange
            dylim = den.yrange
            if dylim > ylim:
                ylim = dylim
            x_coords.append(den.root[0])
        self.den_xlim = den_xlim
        self.den_ylim = ylim
        self.divider = ylim * 1.05

    def draw(
        self,
        ax,
        orient="top",
        spacing=None,
        add_meta=True,
        add_base=True,
        base_colors=None,
        meta_color=None,
        linewidth=None,
        divide=True,
        divide_style="--",
        meta_ratio=0.2,
        rasterized=False,
    ):
        """

        Parameters
        ----------
        ax
        orient
        spacing : float, array of float
            The space between dendrograms
        add_meta : bool
            Draw the meta dendrogram
        add_base : bool
            Draw the base dendrograms
        base_colors : color, array of colors
            The color of base dendrograms, if array is passed,
            will be applied by group order.
        meta_color : color
            The color of meta dendrogram
        linewidth
        divide : bool
            Draw a divide line the divides the meta and base dendrograms
        divide_style :
            The linestyle of the divide line
        meta_ratio : float
            The size of meta dendrogram relative to the base dendrogram
        rasterized : bool
            Rasterize the dendrogram to speed up rendering

        """

        meta_color = ".1" if meta_color is None else meta_color
        linewidth = 0.7 if linewidth is None else linewidth
        if base_colors is None:
            base_colors = cycle([None])
        elif is_color_like(base_colors):
            base_colors = cycle([base_colors])
        else:
            base_colors = np.asarray(base_colors)
            if add_meta:
                base_colors = base_colors[self.reorder_index]

        if spacing is None:
            spacing = [0 for _ in range(self.n - 1)]
        elif not isinstance(spacing, Sequence):
            spacing = [spacing for _ in range(self.n - 1)]

        render_xlim = self.den_xlim / (1 - np.sum(spacing))
        skeleton = np.sort(np.unique(self.x_coords[self.y_coords == 0]))
        ranger = [(skeleton[i], skeleton[i + 1]) for i in range(len(skeleton) - 1)]

        draw_dens = self.dens if add_meta else self.orig_dens

        if add_base:
            x_start = 0
            for i, den in enumerate(draw_dens):
                if x_start != 0:
                    den.set_lim(x_start=x_start, y_end=self.divider)
                else:
                    den.set_lim(y_end=self.divider)
                if i != self.n - 1:
                    x_start = x_start + den.xrange + spacing[i] * render_xlim
            # get render x
            # orient ?
            skeleton_x = [den.render_root[0] for den in draw_dens]

        else:
            xstart = 0
            skeleton_x = []
            for i, den in enumerate(draw_dens):
                if i == 0:
                    xstart += den.xrange / 2
                else:
                    xstart += den.xrange / 2 + spacing[i - 1] * render_xlim
                skeleton_x.append(xstart)
                xstart += den.xrange / 2

        mapper = dict(zip(skeleton, skeleton_x))
        x_coords = np.sort(np.unique(self.x_coords.flatten()))

        real_x = {}

        ranger_ix = 0
        for xc in x_coords:
            render_coord = mapper.get(xc, None)
            if render_coord is None:
                while True:
                    lower, upper = ranger[ranger_ix]
                    real_up, real_low = mapper[upper], mapper[lower]
                    if (xc > lower) & (xc < upper):
                        ratio = (xc - lower) / (upper - lower)
                        real_length = (real_up - real_low) * ratio
                        render_coord = real_low + real_length
                        real_x[xc] = render_coord
                        break
                    else:
                        ranger_ix += 1
            else:
                real_x[xc] = render_coord

        self._render_x_coords = np.asarray(
            [real_x[i] for i in self.x_coords.flatten()]
        ).reshape(self.x_coords.shape)
        if add_base:
            if add_meta:
                norm_y_coords = self.y_coords  # / np.max(self.y_coords)
                amplify = self.den_ylim * meta_ratio
                self._render_y_coords = norm_y_coords * amplify + self.divider
            else:
                self._render_y_coords = self.den_ylim
        else:
            self._render_y_coords = self.y_coords / 5

        if add_meta:
            # Add meta dendrogram
            self._draw_dendrogram(
                ax,
                orient=orient,
                color=meta_color,
                linewidth=linewidth,
                rasterized=rasterized,
            )

        if divide & add_base & add_meta:
            xmin = np.min(draw_dens[0].x_coords)
            xmax = np.max(draw_dens[-1]._render_x_coords)
            if orient in ["top", "bottom"]:
                ax.hlines(
                    self.divider,
                    xmin,
                    xmax,  # 0, xlim,
                    linestyles=divide_style,
                    color=meta_color,
                    linewidth=linewidth,
                    rasterized=rasterized,
                )
            else:
                ax.vlines(
                    self.divider,
                    xmin,
                    xmax,  # 0, ylim,
                    linestyles=divide_style,
                    color=meta_color,
                    linewidth=linewidth,
                    rasterized=rasterized,
                )

        if add_base:
            for den, color in zip(draw_dens, base_colors):
                # The singleton dendrogram will only be drawn if meta is drawn
                if not den.is_singleton or add_meta:
                    den.draw(
                        ax,
                        orient=orient,
                        add_root=add_meta,
                        color=color,
                        linewidth=linewidth,
                        root_color=meta_color,
                        control_ax=False,
                        rasterized=rasterized,
                    )

        xlim = render_xlim
        # reserve room to avoid clipping of the top
        ylim = np.max(self._render_y_coords) * 1.05
        if orient in ["right", "left"]:
            xlim, ylim = ylim, xlim

        ax.set_xlim(0, xlim)
        ax.set_ylim(0, ylim)
        if orient == "left":
            ax.invert_xaxis()
        if orient != "top":
            ax.invert_yaxis()



from typing import Mapping

import numpy as np
import pandas as pd
from legendkit import CatLegend
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.ticker import FuncFormatter
from matplotlib.colors import is_color_like

from ._utils import _format_labels
from .base import StatsBase
from ..utils import ECHARTS16


def simple_bar(
    data,
    ax: Axes = None,
    orient="v",
    width=0.8,
    show_value=True,
    fmt=None,
    label_pad=2,
    text_props=None,
    **kwargs,
):
    if ax is None:
        ax = plt.gca()
    if text_props is None:
        text_props = {}
    bar = ax.bar if orient == "v" else ax.barh
    bars = bar(np.arange(0, len(data)) + 0.5, data, width, **kwargs)
    if show_value:
        ax.bar_label(bars, data, fmt=fmt, padding=label_pad, **text_props)
    return ax


class _BarBase(StatsBase):
    def _process_params(
        self,
        width=0.7,
        orient=None,
        show_value=True,
        fmt=None,
        label=None,
        value_pad=2.0,
        props=None,
        **kwargs,
    ):
        self.width = width
        self.orient = orient
        self.show_value = show_value
        self.label = label
        self.fmt = "%g" if fmt is None else fmt
        self.value_pad = value_pad
        self.props = props if props is not None else {}
        self.options = kwargs


class Numbers(_BarBase):
    """Show numbers in bar plot

    This is used to show 1D data specifically.

    Parameters
    ----------
    data : np.ndarray
        1D data
    width : float
        The width of bar
    color : color or list of color
        The color of bar
    show_value : bool
        Whether to show value on the bar
    fmt : str, callable
        Format the value show on the bar
    label : str
        The label of the plot
    value_pad : float
        The spacing between label and the plot
    props : dict
        See :class:`matplotlib.text.Text`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> from marsilea.plotter import Numbers
        >>> data = np.random.randint(1, 10, 10)
        >>> _, ax = plt.subplots()
        >>> Numbers(data).render(ax)

    """

    def __init__(
        self,
        data,
        width=0.8,
        color="C0",
        orient=None,
        show_value=True,
        fmt=None,
        label=None,
        label_loc=None,
        label_props=None,
        value_pad=2.0,
        props=None,
        **kwargs,
    ):
        self.bars = None

        self._process_params(
            width, orient, show_value, fmt, label, value_pad, props, **kwargs
        )
        self.set_label(label, label_loc, label_props)

        if is_color_like(color):
            color = np.repeat(color, len(data))
        else:
            color = self.data_validator(color, target="1d")
        self.set_data(self.data_validator(data, target="1d"), color)

    def render_ax(self, spec):
        ax = spec.ax
        data, color = spec.data

        lim = len(data)
        orient = self.get_orient()
        bar = ax.bar if orient == "v" else ax.barh
        if orient == "h":
            data = data[::-1]
        self.bars = bar(
            np.arange(0, lim) + 0.5, data, self.width, color=color, **self.options
        )

        if orient == "v":
            ax.set_xlim(0, lim)
        else:
            ax.set_ylim(0, lim)

        if self.side == "left":
            if not ax.xaxis_inverted():
                ax.invert_xaxis()

        if self.show_value:
            ax.bar_label(self.bars, fmt=self.fmt, padding=self.value_pad, **self.props)


class CenterBar(_BarBase):
    """Two comparable bar plots with center line

    Parameters
    ----------
    data : np.ndarray, pd.DataFrame
        If input with array, must be 2D array with shape (n, 2).
        If input with DataFrame, must have two columns.
    names : list of str
        The names of the two bars
    width : float
        The width of bar
    colors : list of color
        The colors of the two bars
    orient : {"v", "h"}
        The orientation of the plot
    show_value : bool
        Whether to show value on the bar
    fmt : str, callable
        Format the value show on the bar
    label : str
        The label of the plot
    value_pad : float
        The spacing between value and the plot
    props : dict
        See :class:`matplotlib.text.Text`
    kwargs:
        Other keyword arguments passed to :meth:`matplotlib.axes.Axes.bar`

    Examples
    --------

    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> from marsilea.plotter import CenterBar
        >>> data = np.random.randint(1, 10, (10, 2))
        >>> _, ax = plt.subplots()
        >>> CenterBar(data, names=["Bar 1", "Bar 2"]).render(ax)


    .. plot::
        :context: close-figs

        >>> _, ax = plt.subplots()
        >>> CenterBar(data, names=["Bar 1", "Bar 2"], orient="h").render(ax)

    """

    def __init__(
        self,
        data,
        names=None,
        width=0.8,
        colors=None,
        orient=None,
        show_value=True,
        fmt=None,
        label=None,
        value_pad=2.0,
        props=None,
        **kwargs,
    ):
        self.set_data(self.data_validator(data.T, target="2d"))
        if names is None:
            if isinstance(data, pd.DataFrame):
                names = data.columns
        self.names = names
        if colors is None:
            colors = ["C0", "C1"]
        self.colors = colors

        self._process_params(
            width, orient, show_value, fmt, label, value_pad, props, **kwargs
        )

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        orient = self.get_orient()
        bar = ax.bar
        line = ax.axhline
        options = {"bottom": 0, **self.options}

        if orient == "h":
            bar = ax.barh
            line = ax.axvline
            data = data[::-1]

            del options["bottom"]
            options["left"] = 0

        left_bar, right_bar = data[0], data[1]
        locs = np.arange(0, len(left_bar)) + 0.5

        bar1 = bar(locs, left_bar, self.width, color=self.colors[0], **self.options)
        bar2 = bar(locs, -right_bar, self.width, color=self.colors[1], **self.options)
        line(0, color="black", lw=1)
        if self.names is not None:
            n1, n2 = self.names
            if orient == "h" and spec.is_first:
                ax.text(0.45, 1, n1, ha="right", va="bottom", transform=ax.transAxes)
                ax.text(0.55, 1, n2, ha="left", va="bottom", transform=ax.transAxes)
            elif orient == "v" and spec.is_last:
                ax.text(1, 0.75, n1, ha="left", va="center", transform=ax.transAxes)
                ax.text(1, 0.25, n2, ha="left", va="center", transform=ax.transAxes)

        lim_value = np.max(data) * 1.05

        if orient == "v":
            ax.set_xlim(0, len(left_bar))
            ax.set_ylim(-lim_value, lim_value)
            ax.yaxis.set_major_formatter(FuncFormatter(lambda x, p: f"{np.abs(x):g}"))
        else:
            ax.set_ylim(0, len(left_bar))
            ax.set_xlim(-lim_value, lim_value)
            ax.xaxis.set_major_formatter(FuncFormatter(lambda x, p: f"{np.abs(x):g}"))

        if self.is_flank:
            ax.invert_yaxis()

        if self.show_value:
            left_label = _format_labels(left_bar, self.fmt)
            right_label = _format_labels(right_bar, self.fmt)
            ax.bar_label(bar1, left_label, padding=self.value_pad, **self.props)
            ax.bar_label(bar2, right_label, padding=self.value_pad, **self.props)


class StackBar(_BarBase):
    """Stacked Bar

    Parameters
    ----------
    data : np.ndarray, pd.DataFrame
        2D data, index of dataframe is used as the name of items.
    items : list of str
        The name of items.
    colors : list of colors, mapping of (item, color)
        The colors of the bar for each item.
    orient : {"v", "h"}
        The orientation of the plot
    show_value : bool
        Whether to show value on the bar
    value_loc : {"center", "edge"}
        The location of the value
    fmt : str, callable
        Format the value show on the bar
    label : str
        The label of the plot
    value_pad : float
        The spacing between value and the bar
    props : dict
        See :class:`matplotlib.text.Text`
    kwargs :
        Other keyword arguments passed to :meth:`matplotlib.axes.Axes.bar`


    Examples
    --------

    .. plot::
        :context: close-figs

        >>> from marsilea.plotter import StackBar
        >>> stack_data = pd.DataFrame(
        ...     data=np.random.randint(1, 10, (5, 10)), index=list("abcde")
        ... )
        >>> _, ax = plt.subplots()
        >>> StackBar(stack_data).render(ax)


    You may find the text is too big for a bar to display on, to not display certain value.

    .. plot::
        :context: close-figs

        >>> fmt = lambda v: int(v) if v > 2 else ""
        >>> _, ax = plt.subplots()
        >>> StackBar(stack_data, show_value=True, fmt=fmt).render(ax)

    """

    def __init__(
        self,
        data,
        items=None,
        colors=None,
        orient=None,
        show_value=False,
        value_loc="center",
        width=0.8,
        value_size=6,
        fmt=None,
        props=None,
        label=None,
        value_pad=0,
        legend_kws=None,
        **kwargs,
    ):
        # TODO: Support bare bone numpy array as input
        item_names = None
        if isinstance(data, pd.DataFrame):
            item_names = data.index
            data = data.to_numpy()
        data = self.data_validator(data, target="2d")

        if items is not None:
            item_names = items

        if colors is None:
            bar_colors = ECHARTS16
        else:
            if isinstance(colors, Mapping):
                if item_names is None:
                    raise ValueError(
                        "Please provide the name of each item "
                        "before assigning color."
                    )
                bar_colors = [colors[name] for name in item_names]
            else:
                bar_colors = colors

        if len(bar_colors) < len(item_names):
            msg = "The number of colors is less than the number of items."
            raise ValueError(msg)

        self.set_data(data)
        self.labels = item_names
        self.bar_colors = bar_colors

        props = {} if props is None else props
        value_props = dict(label_type=value_loc)
        value_props.update(props)

        self._process_params(
            width, orient, show_value, fmt, label, value_pad, value_props, **kwargs
        )

        self.value_size = value_size
        self._legend_kws = dict(title=self.label, size=1)
        if legend_kws is not None:
            self._legend_kws.update(legend_kws)

    def get_legends(self):
        if self.labels is not None:
            return CatLegend(
                colors=self.bar_colors, labels=self.labels, **self._legend_kws
            )

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        orient = self.get_orient()
        bar = ax.bar if orient == "v" else ax.barh

        lim = data.shape[1]
        if orient == "h":
            ax.set_ylim(0, lim)
        else:
            ax.set_xlim(0, lim)
        if self.side == "left":
            if not ax.xaxis_inverted():
                ax.invert_xaxis()

        # Hanlde data
        if orient == "h":
            data = data[:, ::-1]

        locs = np.arange(0, lim) + 0.5
        bottom = np.zeros(lim)

        # reverse to make the order more visually intuitive
        if self.labels is not None:
            labels = self.labels[::-1]
        else:
            labels = [None for _ in range(len(data))]
        colors = self.bar_colors[: len(data)]
        for ix, row in enumerate(data):
            bars = bar(
                locs,
                row,
                self.width,
                bottom,
                fc=colors[ix],
                label=labels[ix],
                **self.options,
            )
            bottom += row

            if self.show_value:
                ax.bar_label(bars, fmt=self.fmt, padding=self.value_pad, **self.props)



"""
Utilities functions to check user input
"""


def check_in_list(options, **kwargs):
    for name, value in kwargs.items():
        if value not in options:
            raise ValueError(
                f"You input unknown {name}={value}, " f"options are {options}"
            )



import numpy as np
from dataclasses import dataclass
from legendkit import cat_legend
from matplotlib.colors import Normalize, is_color_like
from matplotlib.patches import Arc as mArc
from numbers import Number

from .base import StatsBase


@dataclass
class LinkAttrs:
    width: float
    color: str


class Links:
    def __init__(self, links, weights=None, width=None, colors=None, labels=None):
        self._links = {}
        self._links_attr = {}

        self.unweighted = True
        self.width = 1
        self.norm = None
        if weights is not None:
            self.unweighted = False
            self.norm = Normalize()
            self.norm.autoscale(weights)

        if colors is None:
            self.colors = ["C0" for _ in links]
        elif is_color_like(colors):
            self.colors = [colors for _ in links]
        else:
            if len(colors) != len(links):
                msg = (
                    f"Length of colors ({len(colors)}) does not "
                    f"match length of links ({len(links)})"
                )
                raise ValueError(msg)
            self.colors = colors

        self.legend_entries = None
        if labels is not None:
            if len(labels) != len(links):
                msg = (
                    f"Length of labels ({len(labels)}) does not "
                    f"match length of links ({len(links)})"
                )
                raise ValueError(msg)
            self.legend_entries = dict(zip(labels, colors))

        if width is not None:
            if weights is None:
                if isinstance(width, Number):
                    self.width = width
                else:
                    msg = "width must be a number if weights is not specify"
                    raise ValueError(msg)

            else:
                if isinstance(width, tuple):
                    self.width = (np.min(width), np.max(width))
                else:
                    msg = "width must be a tuple of numbers"
                    raise ValueError(msg)
        else:
            if weights is not None:
                self.width = (1, 10)

        if weights is None:
            for (e1, e2), c in zip(links, self.colors):
                self._links.setdefault(e1, []).append(e2)
                self._links_attr[(e1, e2)] = LinkAttrs(self.width, c)
        else:
            w0, w1 = self.width
            for (e1, e2), c, w in zip(links, self.colors, weights):
                self._links.setdefault(e1, []).append(e2)
                width = self.norm(w) * (w1 - w0) + w0
                self._links_attr[(e1, e2)] = LinkAttrs(width, c)

    def get(self, item):
        return self._links.get(item)

    def get_attr(self, link) -> LinkAttrs:
        attr = self._links_attr.get(link)
        if attr is None:
            return self._links_attr.get(link[::-1])
        return attr

    def get_legend_entries(self):
        labels = self.legend_entries.keys()
        colors = self.legend_entries.values()
        return labels, colors


class Arc(StatsBase):
    """Arc diagram.

    Parameters
    ----------
    anchors : array-like
        Anchors of the arcs.
    links : array-like
        Links of the arcs.
    weights : array-like, optional
        Weights of the links. If not specified, all links will have the same
        width.
    width : float or tuple of float, optional
        Width of the links. If `weights` is specified, `width` must be a tuple
        of two floats, where the first float is the minimum width and the
        second float is the maximum width. If `weights` is not specified,
        `width` must be a float.
    colors : str or array-like, optional
        Colors of the links. If not specified, all links will have the same
        color.
    labels : array-like, optional
        Labels of the links. If not specified, no legend will be shown.
    legend_kws : dict, optional
        Keyword arguments passed to `legendkit.cat_legend`.
    label : str, optional
        Label of the plot.
    label_loc : str, optional
        Location of the label. Must be one of the following:
        "left", "right", "top", "bottom", "center".
    label_props : dict, optional
        Keyword arguments passed to `matplotlib.text.Text`.
    **kwargs
        Keyword arguments passed to `matplotlib.patches.Arc`.

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> import marsilea as ma
        >>> from marsilea.plotter import Arc
        >>> anchors = np.arange(10)
        >>> links = [(0, 7), (1, 5), (3, 10), (0, 9), (5, 8), (2, 6), (4, 8)]
        >>> weights = [1, 2, 3, 5, 2, 1, 2]
        >>> colors = ["C0", "C1", "C2", "C3", "C4", "C5", "C6"]
        >>> labels = ["A", "B", "C", "D", "E", "F", "G"]
        >>> h = ma.Heatmap(np.random.rand(10, 10))
        >>> h.add_top(Arc(anchors, links, weights=weights, colors=colors, labels=labels))
        >>> h.render()



    """

    render_main = False

    def __init__(
        self,
        anchors,
        links,
        weights=None,
        width=None,
        colors=None,
        labels=None,
        legend_kws=None,
        label=None,
        label_loc=None,
        label_props=None,
        **kwargs,
    ):
        if len(np.unique(anchors)) != len(anchors):
            raise ValueError("`anchors` must be unique")
        anchors = self.data_validator(anchors, target="1d")
        self.set_data(anchors)
        self.links = Links(
            links, weights=weights, width=width, colors=colors, labels=labels
        )
        self.options = kwargs
        self.legend_kws = {} if legend_kws is None else legend_kws

        self.set_label(label, label_loc, label_props)

    def render_ax(self, spec):
        ax = spec.ax
        data = spec.data

        chunks = np.linspace(0, 1, len(data) + 1)
        mids = (chunks[1:] + chunks[:-1]) / 2
        anchors_coords = dict(zip(data, mids))

        sizes = []

        for anchor in data:
            # is this anchor has a link?
            links = self.links.get(anchor)
            if links is not None:
                # is this link exist in axes?
                for link in links:
                    link_in_data = anchors_coords.get(link)
                    if link_in_data is not None:
                        arc_start = anchors_coords[anchor]
                        arc_end = link_in_data
                        if arc_end < arc_start:
                            arc_start, arc_end = arc_end, arc_start
                        arc_mid = (arc_start + arc_end) / 2
                        arc_width = arc_end - arc_start

                        link_attr = self.links.get_attr((anchor, link))
                        options = dict(
                            theta1=0,
                            theta2=180,
                            color=link_attr.color,
                            linewidth=link_attr.width,
                            **self.options,
                        )

                        if self.side == "left":
                            xy = (0, arc_mid)
                            angle = -90
                        elif self.side == "right":
                            xy = (0, arc_mid)
                            angle = -90
                        else:
                            xy = (arc_mid, 0)
                            angle = 0
                        sizes.append(arc_width)
                        arc = mArc(xy, arc_width, arc_width * 2, angle=angle, **options)
                        ax.add_patch(arc)

        lim = np.max(sizes)
        if self.is_flank:
            ax.set_xlim(0, lim * 1.1)
            ax.set_ylim(1, 0)
        else:
            ax.set_ylim(lim * 1.1, 0)
            ax.set_xlim(0, 1)
        if self.side == "top":
            ax.invert_yaxis()
        if self.side == "left":
            ax.invert_xaxis()
        ax.set_axis_off()

    def get_legends(self):
        if self.links.legend_entries is not None:
            labels, colors = self.links.get_legend_entries()
            return cat_legend(labels=labels, colors=colors, **self.legend_kws)
